import base64
import http.server
import ipaddress
import os
import signal
import sys
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer
from socketserver import ThreadingMixIn

from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.exc import NoResultFound

from keylime import api_version as keylime_api_version
from keylime import cert_utils, crypto, json, keylime_logging, web_util
from keylime.common import validators
from keylime.db.keylime_db import DBEngineManager, SessionManager
from keylime.db.registrar_db import RegistrarMain
from keylime.tpm import tpm2_objects
from keylime.tpm.tpm_main import tpm

logger = keylime_logging.init_logging("registrar")


try:
    engine = DBEngineManager().make_engine("registrar")
except SQLAlchemyError as err:
    logger.error("Error creating SQL engine: %s", err)
    sys.exit(1)


class ProtectedHandler(BaseHTTPRequestHandler, SessionManager):
    def do_HEAD(self):
        """HEAD not supported"""
        web_util.echo_json_response(self, 405, "HEAD not supported")

    def do_PATCH(self):
        """PATCH not supported"""
        web_util.echo_json_response(self, 405, "PATCH not supported")

    def do_GET(self):
        """This method handles the GET requests to retrieve status on agents from the Registrar Server.

        Currently, only agents resources are available for GETing, i.e. /agents. All other GET uri's
        will return errors. agents requests require a single agent_id parameter which identifies the
        agent to be returned. If the agent_id is not found, a 404 response is returned.
        """
        session = SessionManager().make_session(engine)
        rest_params = web_util.get_restful_params(self.path)
        if rest_params is None:
            web_util.echo_json_response(self, 405, "Not Implemented: Use /agents/ interface")
            return

        if not web_util.validate_api_version(self, rest_params["api_version"], logger):
            return

        if "agents" not in rest_params:
            web_util.echo_json_response(self, 400, "uri not supported")
            logger.warning("GET returning 400 response. uri not supported: %s", self.path)
            return

        agent_id = rest_params["agents"]

        if agent_id is not None:
            # If the agent ID is not valid (wrong set of characters),
            # just do nothing.
            if not validators.valid_agent_id(agent_id):
                web_util.echo_json_response(self, 400, "agent_id not not valid")
                logger.error("GET received an invalid agent ID: %s", agent_id)
                return

            agent = None
            try:
                agent = session.query(RegistrarMain).filter_by(agent_id=agent_id).first()
            except SQLAlchemyError as e:
                logger.error("SQLAlchemy Error for agent ID %s: %s", agent_id, e)
                return

            if agent is None:
                web_util.echo_json_response(self, 404, f"agent {agent_id} not found")
                logger.warning("GET returning 404 response. agent %s not found.", agent_id)
                return

            if not bool(agent.active):
                web_util.echo_json_response(self, 404, f"agent {agent_id} not yet active")
                logger.warning("GET returning 404 response. agent %s not yet active.", agent_id)
                return

            response = {
                "aik_tpm": agent.aik_tpm,
                "ek_tpm": agent.ek_tpm,
                "ekcert": agent.ekcert,
                "mtls_cert": agent.mtls_cert,
                "ip": agent.ip,
                "port": agent.port,
                "regcount": agent.regcount,
            }

            if agent.virtual:
                response["provider_keys"] = agent.provider_keys

            web_util.echo_json_response(self, 200, "Success", response)
            logger.info("GET returning 200 response for agent_id: %s", agent_id)
        else:
            # return the available registered uuids from the DB
            json_response = session.query(RegistrarMain.agent_id).all()
            return_response = [item[0] for item in json_response]
            web_util.echo_json_response(self, 200, "Success", {"uuids": return_response})
            logger.info("GET returning 200 response for agent_id list")

        return

    def do_POST(self):
        """POST not supported"""
        web_util.echo_json_response(self, 405, "POST not supported via TLS interface")

    def do_PUT(self):
        """PUT not supported"""
        web_util.echo_json_response(self, 405, "PUT not supported via TLS interface")

    def do_DELETE(self):
        """This method handles the DELETE requests to remove agents from the Registrar Server.

        Currently, only agents resources are available for DELETEing, i.e. /agents. All other DELETE uri's will return errors.
        agents requests require a single agent_id parameter which identifies the agent to be deleted.
        """
        session = SessionManager().make_session(engine)
        rest_params = web_util.get_restful_params(self.path)
        if rest_params is None:
            web_util.echo_json_response(self, 405, "Not Implemented: Use /agents/ interface")
            return

        if not web_util.validate_api_version(self, rest_params["api_version"], logger):
            return

        if "agents" not in rest_params:
            web_util.echo_json_response(self, 400, "URI not supported")
            logger.warning("DELETE agent returning 400 response. uri not supported: %s", self.path)
            return

        agent_id = rest_params["agents"]

        if agent_id is not None:
            # If the agent ID is not valid (wrong set of characters),
            # just do nothing.
            if not validators.valid_agent_id(agent_id):
                web_util.echo_json_response(self, 400, "agent_id not not valid")
                logger.error("DELETE received an invalid agent ID: %s", agent_id)
                return

            if session.query(RegistrarMain).filter_by(agent_id=agent_id).delete():
                # send response
                try:
                    session.commit()
                except SQLAlchemyError as e:
                    logger.error("SQLAlchemy Error: %s", e)
                web_util.echo_json_response(self, 200, "Success")
                return

            # send response
            web_util.echo_json_response(self, 404)
            return

        web_util.echo_json_response(self, 404)

    # pylint: disable=W0622
    def log_message(self, format, *args):
        return


class UnprotectedHandler(BaseHTTPRequestHandler, SessionManager):
    def do_HEAD(self):
        """HEAD not supported"""
        web_util.echo_json_response(self, 405, "HEAD not supported")

    def do_PATCH(self):
        """PATCH not supported"""
        web_util.echo_json_response(self, 405, "PATCH not supported")

    def do_GET(self):
        """This method handles the GET requests to the unprotected side of the Registrar Server

        Currently the only supported path is /versions which shows the supported API versions
        """
        rest_params = web_util.get_restful_params(self.path)
        if rest_params is None:
            web_util.echo_json_response(self, 405, "Not Implemented: Use /version/ interface")
            return

        if "version" not in rest_params:
            web_util.echo_json_response(self, 400, "URI not supported")
            logger.warning("GET agent returning 400 response. URI not supported: %s", self.path)
            return

        version_info = {
            "current_version": keylime_api_version.current_version(),
            "supported_versions": keylime_api_version.all_versions(),
        }

        web_util.echo_json_response(self, 200, "Success", version_info)

    def do_POST(self):
        """This method handles the POST requests to add agents to the Registrar Server.

        Currently, only agents resources are available for POSTing, i.e. /agents. All other POST uri's
        will return errors. POST requests require an an agent_id identifying the agent to add, and json
        block sent in the body with 2 entries: ek and aik.
        """
        session = SessionManager().make_session(engine)
        rest_params = web_util.get_restful_params(self.path)
        if rest_params is None:
            web_util.echo_json_response(self, 405, "Not Implemented: Use /agents/ interface")
            return

        if not web_util.validate_api_version(self, rest_params["api_version"], logger):
            return

        if "agents" not in rest_params:
            web_util.echo_json_response(self, 400, "uri not supported")
            logger.warning("POST agent returning 400 response. uri not supported: %s", self.path)
            return

        agent_id = rest_params["agents"]

        if agent_id is None:
            web_util.echo_json_response(self, 400, "agent id not found in uri")
            logger.warning("POST agent returning 400 response. agent id not found in uri %s", self.path)
            return

        # If the agent ID is not valid (wrong set of characters), just
        # do nothing.
        if not validators.valid_agent_id(agent_id):
            web_util.echo_json_response(self, 400, "agent id not valid")
            logger.error("POST received an invalid agent ID: %s", agent_id)
            return

        try:
            content_length = int(self.headers.get("Content-Length", 0))
            if content_length == 0:
                web_util.echo_json_response(self, 400, "Expected non zero content length")
                logger.warning("POST for %s returning 400 response. Expected non zero content length.", agent_id)
                return

            post_body = self.rfile.read(content_length)
            json_body = json.loads(post_body)

            ekcert = json_body["ekcert"]
            aik_tpm = json_body["aik_tpm"]

            initialize_tpm = tpm()

            if ekcert is None or ekcert == "emulator":
                logger.warning("Agent %s did not submit an ekcert", agent_id)
                ek_tpm = json_body["ek_tpm"]
            else:
                if "ek_tpm" in json_body:
                    # This would mean the agent submitted both a non-None ekcert, *and*
                    #  an ek_tpm... We can deal with it by just ignoring the ek_tpm they sent
                    logger.warning("Overriding ek_tpm for agent %s from ekcert", agent_id)
                # If there's an EKCert, we just overwrite their ek_tpm
                # Note, we don't validate the EKCert here, other than the implicit
                #  "is it a valid x509 cert" check. So it's still untrusted.
                # This will be validated by the tenant.
                cert = cert_utils.x509_der_cert(base64.b64decode(ekcert))
                ek_tpm = base64.b64encode(tpm2_objects.ek_low_tpm2b_public_from_pubkey(cert.public_key())).decode()

            aik_attrs = tpm2_objects.get_tpm2b_public_object_attributes(
                base64.b64decode(aik_tpm),
            )
            if aik_attrs != tpm2_objects.AK_EXPECTED_ATTRS:
                web_util.echo_json_response(self, 400, "Invalid AK attributes")
                logger.warning(
                    "Agent %s submitted AIK with invalid attributes! %s (provided) != %s (expected)",
                    agent_id,
                    tpm2_objects.object_attributes_description(aik_attrs),
                    tpm2_objects.object_attributes_description(tpm2_objects.AK_EXPECTED_ATTRS),
                )
                return

            # try to encrypt the AIK
            (blob, key) = initialize_tpm.encryptAIK(
                agent_id,
                base64.b64decode(ek_tpm),
                base64.b64decode(aik_tpm),
            )

            # special behavior if we've registered this uuid before
            regcount = 1
            try:
                agent = session.query(RegistrarMain).filter_by(agent_id=agent_id).first()
            except NoResultFound:
                agent = None
            except SQLAlchemyError as e:
                logger.error("SQLAlchemy Error: %s", e)
                raise

            if agent is not None:

                # keep track of how many ek-ekcerts have registered on this uuid
                regcount = agent.regcount
                if agent.ek_tpm != ek_tpm or agent.ekcert != ekcert:
                    logger.warning("WARNING: Overwriting previous registration for this UUID with new ek-ekcert pair!")
                    regcount += 1

                # force overwrite
                logger.info("Overwriting previous registration for this UUID.")
                try:
                    session.query(RegistrarMain).filter_by(agent_id=agent_id).delete()
                    session.commit()
                except SQLAlchemyError as e:
                    logger.error("SQLAlchemy Error: %s", e)
                    raise
            # Check for ip and port
            contact_ip = json_body.get("ip", None)
            contact_port = json_body.get("port", None)

            # Validate ip and port
            if contact_ip is not None:
                try:
                    # Use parser from the standard library instead of implementing our own
                    ipaddress.ip_address(contact_ip)
                except ValueError:
                    logger.warning("Contact ip for agent %s is not a valid ip got: %s.", agent_id, contact_ip)
                    contact_ip = None
            if contact_port is not None:
                try:
                    contact_port = int(contact_port)
                    if contact_port < 1 or contact_port > 65535:
                        logger.warning(
                            "Contact port for agent %s is not a number between 1 and got: %s.", agent_id, contact_port
                        )
                        contact_port = None
                except ValueError:
                    logger.warning("Contact port for agent %s is not a valid number got: %s.", agent_id, contact_port)
                    contact_port = None

            # Check for mTLS cert
            mtls_cert = json_body.get("mtls_cert", None)
            if mtls_cert is None or mtls_cert == "disabled":
                logger.warning("Agent %s did not send a mTLS certificate. Most operations will not work!", agent_id)

            # Add values to database
            d = {}
            d["agent_id"] = agent_id
            d["ek_tpm"] = ek_tpm
            d["aik_tpm"] = aik_tpm
            d["ekcert"] = ekcert
            d["ip"] = contact_ip
            d["mtls_cert"] = mtls_cert
            d["port"] = contact_port
            d["virtual"] = int(ekcert == "virtual")
            d["active"] = int(False)
            d["key"] = key
            d["provider_keys"] = {}
            d["regcount"] = regcount

            try:
                session.add(RegistrarMain(**d))
                session.commit()
            except SQLAlchemyError as e:
                logger.error("SQLAlchemy Error: %s", e)
                raise

            response = {
                "blob": blob,
            }
            web_util.echo_json_response(self, 200, "Success", response)

            logger.info("POST returning key blob for agent_id: %s", agent_id)
        except Exception as e:
            web_util.echo_json_response(self, 400, f"Error: {str(e)}")
            logger.warning("POST for %s returning 400 response. Error: %s", agent_id, e)
            logger.exception(e)

    def do_PUT(self):
        """This method handles the PUT requests to add agents to the Registrar Server.

        Currently, only agents resources are available for PUTing, i.e. /agents. All other PUT uri's
        will return errors.
        """
        session = SessionManager().make_session(engine)
        rest_params = web_util.get_restful_params(self.path)
        if rest_params is None:
            web_util.echo_json_response(self, 405, "Not Implemented: Use /agents/ interface")
            return

        if not web_util.validate_api_version(self, rest_params["api_version"], logger):
            return

        if "agents" not in rest_params:
            web_util.echo_json_response(self, 400, "uri not supported")
            logger.warning("PUT agent returning 400 response. uri not supported: %s", self.path)
            return

        agent_id = rest_params["agents"]

        if agent_id is None:
            web_util.echo_json_response(self, 400, "agent id not found in uri")
            logger.warning("PUT agent returning 400 response. agent id not found in uri %s", self.path)
            return

        # If the agent ID is not valid (wrong set of characters), just
        # do nothing.
        if not validators.valid_agent_id(agent_id):
            web_util.echo_json_response(self, 400, "agent_id not not valid")
            logger.error("PUT received an invalid agent ID: %s", agent_id)
            return

        try:
            content_length = int(self.headers.get("Content-Length", 0))
            if content_length == 0:
                web_util.echo_json_response(self, 400, "Expected non zero content length")
                logger.warning("PUT for %s returning 400 response. Expected non zero content length.", agent_id)
                return

            post_body = self.rfile.read(content_length)
            json_body = json.loads(post_body)

            auth_tag = json_body["auth_tag"]
            try:
                agent = session.query(RegistrarMain).filter_by(agent_id=agent_id).first()
            except NoResultFound as e:
                raise Exception("attempting to activate agent before requesting " f"registrar for {agent_id}") from e
            except SQLAlchemyError as e:
                logger.error("SQLAlchemy Error: %s", e)
                raise

            ex_mac = crypto.do_hmac(agent.key.encode(), agent_id)
            if ex_mac == auth_tag:
                try:
                    session.query(RegistrarMain).filter(RegistrarMain.agent_id == agent_id).update(
                        {"active": int(True)}
                    )
                    session.commit()
                except SQLAlchemyError as e:
                    logger.error("SQLAlchemy Error: %s", e)
                    raise
            else:
                raise Exception(f"Auth tag {auth_tag} does not match expected value {ex_mac}")

            web_util.echo_json_response(self, 200, "Success")
            logger.info("PUT activated: %s", agent_id)
        except Exception as e:
            web_util.echo_json_response(self, 400, f"Error: {str(e)}")
            logger.warning("PUT for %s returning 400 response. Error: %s", agent_id, e)
            logger.exception(e)
            return

    def do_DELETE(self):
        """DELETE not supported"""
        web_util.echo_json_response(self, 405, "DELETE not supported")

    # pylint: disable=W0622
    def log_message(self, format, *args):
        return


# consider using PooledProcessMixIn
# https://github.com/muayyad-alsadi/python-PooledProcessMixIn


class RegistrarServer(ThreadingMixIn, HTTPServer):
    """Handle requests in a separate thread."""

    def __init__(self, server_address, RequestHandlerClass):
        """Constructor overridden to provide ability to read file"""
        http.server.HTTPServer.__init__(self, server_address, RequestHandlerClass)

    def shutdown(self):
        http.server.HTTPServer.shutdown(self)


def start(host, tlsport, port):
    """Main method of the Registrar Server.  This method is encapsulated in a function for packaging to allow it to be
    called as a function by an external program."""

    # set a conservative general umask
    os.umask(0o077)

    RegistrarMain.metadata.create_all(engine, checkfirst=True)
    session = SessionManager().make_session(engine)
    try:
        count = session.query(RegistrarMain.agent_id).count()
        if count > 0:
            logger.info("Loaded %d public keys from database", count)
    except SQLAlchemyError as e:
        logger.error("SQLAlchemy Error: %s", e)

    # Set up the protected registrar server
    protected_server = RegistrarServer((host, tlsport), ProtectedHandler)
    context = web_util.init_mtls("registrar", logger=logger)
    if context is not None:
        protected_server.socket = context.wrap_socket(protected_server.socket, server_side=True)
    thread_protected_server = threading.Thread(target=protected_server.serve_forever)

    # Set up the unprotected registrar server
    unprotected_server = RegistrarServer((host, port), UnprotectedHandler)
    thread_unprotected_server = threading.Thread(target=unprotected_server.serve_forever)

    logger.info("Starting Cloud Registrar Server on ports %s and %s (TLS) use <Ctrl-C> to stop", port, tlsport)
    keylime_api_version.log_api_versions(logger)
    thread_protected_server.start()
    thread_unprotected_server.start()

    def signal_handler(signum, frame):
        del signum, frame
        logger.info("Shutting down Registrar Server...")
        protected_server.shutdown()
        unprotected_server.shutdown()
        sys.exit(0)

    # Catch these signals.  Note that a SIGKILL cannot be caught, so
    # killing this process with "kill -9" may result in improper shutdown
    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGQUIT, signal_handler)
    signal.signal(signal.SIGINT, signal_handler)

    thread_protected_server.join()
    thread_unprotected_server.join()



import argparse
import copy
import datetime
import json
import sys
from os.path import basename

from keylime.ima import file_signatures, ima

# pylint: disable=pointless-string-statement
"""
This script converts legacy Keylime policies into the unified, next-generation
IMA policy format. Both legacy allow/exclude lists as well as out-of-date IMA
policy formats are accepted.

Example usage:

To convert an allowlist and exclude list to an IMA policy at `keylime-policy.json`:

```
python convert_ima_policy.py -a <allowlist path> -e <exclude list path> -o keylime-policy.json
```

To convert an existing IMA policy to the latest version:

```
python convert_ima_policy.py -i <ima policy path> -o keylime-policy.json
```

To view help and see all available options:

```
python convert_ima_policy.py -h
```
"""

# Creates an IMA policy from provided legacy allowlist.
def convert_legacy_allowlist(allowlist_path):
    with open(allowlist_path, "r", encoding="utf8") as f:
        alist_raw = f.read()

    # Attempt to load JSON, and convert the appropriate format
    try:
        alist_json = json.loads(alist_raw)
        print(f"{basename(allowlist_path)} appears to be JSON-formatted; converting to IMA policy")
        ima_policy = _convert_json_allowlist(alist_json)
    except Exception as _:
        print(
            f"{basename(allowlist_path)} is not JSON-formatted; attempting to convert to IMA policy from flat file format"
        )
        ima_policy = _convert_flat_format_allowlist(alist_raw)
    return ima_policy


# Converts JSON-format allowlist to JSON-format IMA policy
def _convert_json_allowlist(alist_json):
    ima_policy = copy.deepcopy(ima.EMPTY_IMA_POLICY)
    ima_policy["meta"]["timestamp"] = str(datetime.datetime.now())
    ima_policy["meta"]["generator"] = "convert_ima_policy.py"
    for key in ima_policy.keys():
        if key == "digests":
            digests = alist_json.get("hashes")
            if not digests:
                print("Allowlist does not have a valid hash list!")
            else:
                ima_policy[key] = alist_json["hashes"]
        elif key == "meta":
            # Skip old metadata
            continue
        else:
            to_migrate = alist_json.get(key)
            if not to_migrate:
                print(f"IMA policy field '{key}' not found in allowlist; using default value")
            else:
                ima_policy[key] = alist_json[key]
    return ima_policy


# Converts flat-format allowlist to JSON-format IMA policy
def _convert_flat_format_allowlist(alist_raw):
    ima_policy = copy.deepcopy(ima.EMPTY_IMA_POLICY)
    ima_policy["meta"]["timestamp"] = str(datetime.datetime.now())
    ima_policy["meta"]["generator"] = "convert-ima-policy"

    lines = alist_raw.splitlines()
    for line_num, line in enumerate(lines):
        line = line.strip()
        if len(line) == 0:
            continue

        pieces = line.split(None, 1)
        if not len(pieces) == 2:
            print(f"Line #{line_num} in Allowlist does not consist of hash and file path: {line}")
            continue

        (checksum_hash, path) = pieces

        if path.startswith("%keyring:"):
            entrytype = "keyrings"
            path = path[len("%keyring:") :]  # remove leading '%keyring:' from path to get keyring name
        else:
            entrytype = "digests"

        if path in ima_policy[entrytype]:
            ima_policy[entrytype][path].append(checksum_hash)
        else:
            ima_policy[entrytype][path] = [checksum_hash]
    return ima_policy


# Updates an existing IMA policy to the latest version, and adds any provided input
def update_ima_policy(policy, excludelist_path=None, verification_keys=None):
    if policy["meta"]["version"] < ima.IMA_POLICY_CURRENT_VERSION:
        print(
            f"Provided policy has version {policy['meta']['version']}; latest policy has version {ima.IMA_POLICY_CURRENT_VERSION}. Updating to latest version."
        )
        updated_policy = copy.deepcopy(ima.IMA_POLICY_CURRENT_VERSION)
        updated_policy["meta"]["timestamp"] = str(datetime.datetime.now())
        updated_policy["meta"]["generator"] = "convert_ima_policy.py"
        for key in updated_policy.keys():
            if key == "meta":
                continue
            to_migrate = policy.get(key)
            if not to_migrate:
                print(f"IMA policy field '{key}' not found in existing IMA policy; using default value")
            else:
                updated_policy[key] = policy[key]
        policy = updated_policy

    excl_list = []
    if excludelist_path:
        with open(excludelist_path, "r", encoding="utf8") as f:
            for line in f:
                line = line.strip()
                if line.startswith("#") or len(line) == 0:
                    continue
                excl_list.append(line)

    verification_key_list = None
    if verification_keys:
        keyring = file_signatures.ImaKeyring().from_string(policy["ima_sign_verification_keys"])
        for key in verification_keys:
            pubkey, keyidv2 = file_signatures.get_pubkey_from_file(key)
            if not pubkey:
                print(f"File '{key}' is not a file with a key")
            keyring.add_pubkey(pubkey, keyidv2)
        verification_key_list = json.dumps(keyring.to_string())

    policy["excludes"] += excl_list
    if verification_key_list:
        policy["verification-keys"].append(verification_key_list)

    return policy


def main():

    parser = ConversionParser()
    parser.add_argument("-a", "--allowlist", help="allowlist file location", action="store")
    parser.add_argument("-i", "--ima_policy", help="IMA policy file location", action="store")
    parser.add_argument("-e", "--excludelist", help="exclude list file location", action="store")
    parser.add_argument("-v", "--verification_keys", help="list of verification key paths", action="store", type=list)
    parser.add_argument("-o", "--output_file", help="Output file path", action="store")

    # print help if no arguments provided
    if len(sys.argv) == 1:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    if (bool(args.allowlist) and not bool(args.ima_policy)) and (not bool(args.allowlist) and bool(args.ima_policy)):
        print("Either --allowlist or --ima_policy is required!")
        sys.exit(1)
    elif not args.output_file:
        print("An output file path (-o, --output_file) is required to write new policy!")
        sys.exit(1)

    if args.allowlist:
        policy = convert_legacy_allowlist(args.allowlist)
    elif args.ima_policy:
        with open(args.ima_policy, "r", encoding="utf8") as f:
            policy = json.load(f)
    policy_out = update_ima_policy(policy, excludelist_path=args.excludelist, verification_keys=args.verification_keys)
    with open(args.output_file, "wb") as f:
        f.write(json.dumps(policy_out).encode())


class ConversionParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write(f"error: {message}\n")
        self.print_help(sys.stderr)
        sys.exit(2)


if __name__ == "__main__":
    main()



import os
import shutil

from keylime import cmd_exec, config, keylime_logging

logger = keylime_logging.init_logging("secure_mount")

# Store the mounted directories done by Keylime, so we can unmount
# them in reverse order
_MOUNTED = []


def check_mounted(secdir):
    """Inspect mountinfo to detect if a directory is mounted."""
    secdir_escaped = secdir.replace(" ", r"\040")
    with open("/proc/self/mountinfo", "r", encoding="utf-8") as f:
        # because of tests we use readlines() and avoid using iterator
        # since mocked open cannot use iterator on Python 3.6
        # see https://code-examples.net/en/q/17a1c75
        #     https://bugs.python.org/issue21258
        for line in f.readlines():
            # /proc/[pid]/mountinfo have 10+ elements separated with
            # spaces (check proc (5) for a complete description)
            #
            # At position 7 there are some optional fields, so we need
            # first to determine the separator mark, and validate the
            # final total number of fields.
            elements = line.split()
            try:
                separator = elements.index("-")
            except ValueError:
                msg = "Separator field not found. Information line cannot be parsed"
                logger.error(msg)
                # pylint: disable=raise-missing-from
                raise Exception(msg)

            if len(elements) < 10 or len(elements) - separator < 4:
                msg = "Mount information line cannot be parsed"
                logger.error(msg)
                raise Exception(msg)

            mount_point = elements[4]
            filesystem_type = elements[separator + 1]
            if mount_point == secdir_escaped:
                if filesystem_type != "tmpfs":
                    msg = (
                        f"Secure storage location {secdir} already mounted "
                        f"on wrong file system type: {filesystem_type}. "
                        "Unmount to continue."
                    )
                    logger.error(msg)
                    raise Exception(msg)

                logger.debug("Secure storage location %s already mounted on tmpfs", secdir)
                return True

    logger.debug("Secure storage location %s not mounted", secdir)
    return False


def get_secdir():
    secdir = os.path.join(config.WORK_DIR, "secure")

    if not config.MOUNT_SECURE:
        secdir = os.path.join(config.WORK_DIR, "tmpfs-dev")

    return secdir


def mount():
    secdir = get_secdir()

    if not config.MOUNT_SECURE:
        if not os.path.isdir(secdir):
            os.makedirs(secdir)
        return secdir

    if not check_mounted(secdir):
        # ok now we know it isn't already mounted, go ahead and create and mount
        if not os.path.exists(secdir):
            os.makedirs(secdir, 0o700)
        size = config.get("agent", "secure_size")
        logger.info("mounting secure storage location %s on tmpfs", secdir)
        cmd = ("mount", "-t", "tmpfs", "-o", f"size={size},mode=0700", "tmpfs", secdir)
        cmd_exec.run(cmd)
        _MOUNTED.append(secdir)

    return secdir


def umount():
    """Umount all the devices mounted by Keylime."""

    # Make sure we leave tmpfs dir empty even if we did not mount it or
    # if we cannot unmount it. Ignore errors while deleting. The deletion
    # of the 'secure' directory will result in an error since it's a mount point.
    # Also, with config.MOUNT_SECURE being False we remove the directory
    secdir = get_secdir()
    if not config.MOUNT_SECURE or check_mounted(secdir):
        shutil.rmtree(secdir, ignore_errors=True)

    while _MOUNTED:
        directory = _MOUNTED.pop()
        logger.info("Unmounting %s", directory)
        if check_mounted(directory):
            cmd = ("umount", directory)
            ret = cmd_exec.run(cmd, raiseOnError=False)
            if ret["code"] != 0:
                logger.error(
                    "%s cannot be umounted. A running process can be keeping it bussy: %s",
                    directory,
                    str(ret["reterr"]),
                )
        else:
            logger.warning("%s already unmounted by another process", directory)



import io
import os.path
import subprocess
import sys

from cryptography import exceptions as crypto_exceptions
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec, padding
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from pyasn1.codec.der import decoder, encoder
from pyasn1_modules import pem, rfc2459

from keylime import config, keylime_logging, tpm_ek_ca

# Issue #944 -- python-cryptography won't parse malformed certs,
# such as some Nuvoton ones we have encountered in the field.
# Unfortunately, we still have to deal with such certs anyway.

# Here we provide some helpers that use pyasn1 to parse the certificates
# when parsing them with python-cryptography fails, and in this case, we
# try to read the parsed certificate again into python-cryptograhy.

logger = keylime_logging.init_logging("cert_utils")


def x509_der_cert(der_cert_data: bytes):
    """Load an x509 certificate provided in DER format
    :param der_cert_data: the DER bytes of the certificate
    :type der_cert_data: bytes
    :returns: cryptography.x509.Certificate
    """
    try:
        return x509.load_der_x509_certificate(data=der_cert_data, backend=default_backend())
    except Exception as e:
        logger.warning("Failed to parse DER data with python-cryptography: %s", e)
        pyasn1_cert = decoder.decode(der_cert_data, asn1Spec=rfc2459.Certificate())[0]
        return x509.load_der_x509_certificate(data=encoder.encode(pyasn1_cert), backend=default_backend())


def x509_pem_cert(pem_cert_data: str):
    """Load an x509 certificate provided in PEM format
    :param pem_cert_data: the base-64 encoded PEM certificate
    :type pem_cert_data: str
    :returns: cryptography.x509.Certificate
    """
    try:
        return x509.load_pem_x509_certificate(data=pem_cert_data.encode("utf-8"), backend=default_backend())
    except Exception as e:
        logger.warning("Failed to parse PEM data with python-cryptography: %s", e)
        # Let's read the DER bytes from the base-64 PEM.
        der_data = pem.readPemFromFile(io.StringIO(pem_cert_data))
        # Now we can load it as we do in x509_der_cert().
        pyasn1_cert = decoder.decode(der_data, asn1Spec=rfc2459.Certificate())[0]
        return x509.load_der_x509_certificate(data=encoder.encode(pyasn1_cert), backend=default_backend())


def verify_ek(ekcert, tpm_cert_store=config.get("tenant", "tpm_cert_store")):
    """Verify that the provided EK certificate is signed by a trusted root
    :param ekcert: The Endorsement Key certificate in DER format
    :returns: True if the certificate can be verified, False otherwise
    """
    try:
        trusted_certs = tpm_ek_ca.cert_loader(tpm_cert_store)
    except Exception as e:
        logger.warning("Error loading trusted certificates from the TPM cert store: %s", e)
        return False

    try:
        ek509 = x509_der_cert(ekcert)
        for cert_file, pem_cert in trusted_certs.items():
            signcert = x509_pem_cert(pem_cert)
            if ek509.issuer != signcert.subject:
                continue

            signcert_pubkey = signcert.public_key()
            try:
                if isinstance(signcert_pubkey, RSAPublicKey):
                    signcert_pubkey.verify(
                        ek509.signature,
                        ek509.tbs_certificate_bytes,
                        padding.PKCS1v15(),
                        ek509.signature_hash_algorithm,
                    )
                elif isinstance(signcert_pubkey, EllipticCurvePublicKey):
                    signcert_pubkey.verify(
                        ek509.signature,
                        ek509.tbs_certificate_bytes,
                        ec.ECDSA(ek509.signature_hash_algorithm),
                    )
                else:
                    logger.warning("Unsupported public key type: %s", type(signcert_pubkey))
                    continue
            except crypto_exceptions.InvalidSignature:
                continue

            logger.debug("EK cert matched cert: %s", cert_file)
            return True
    except Exception as e:
        # Log the exception so we don't lose the raw message
        logger.exception(e)
        raise Exception("Error processing ek/ekcert. Does this TPM have a valid EK?").with_traceback(sys.exc_info()[2])

    logger.error("No Root CA matched EK Certificate")
    return False


def verify_ek_script(script, env, cwd):
    if script is None:
        logger.warning("External check script (%s) not specified", script)
        return False

    script_path = os.path.abspath(script)
    if not os.path.isfile(script_path):
        if cwd is None or not os.path.isfile(os.path.abspath(os.path.join(cwd, script))):
            logger.warning("External check script (%s) not found; please make sure its path is correct", script)
            return False
        script_path = os.path.abspath(os.path.join(cwd, script))

    try:
        proc = subprocess.run(
            [script_path],
            env=env,
            shell=False,
            cwd=cwd,
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            check=False,
        )
        if proc.returncode != 0:
            errmsg = ""
            if proc.stdout is not None:
                errmsg = proc.stdout.decode("utf-8")
            logger.error("External check script failed to validate EK: %s", errmsg)
            return False
        logger.debug("External check script successfully to validated EK")
        if proc.stdout is not None:
            logger.info("ek_check output: %s", proc.stdout.decode("utf-8"))
    except subprocess.CalledProcessError as e:
        logger.error("Error while trying to run external check script to validate EK: %s", e)
        return False
    return True



import glob
import os

from keylime import config, keylime_logging

logger = keylime_logging.init_logging("tpm_ek_ca")
trusted_certs = {}


def check_tpm_cert_store(tpm_cert_store=config.get("tenant", "tpm_cert_store")):
    if not os.path.isdir(tpm_cert_store):
        logger.error("The directory %s does not exist.", tpm_cert_store)
        raise Exception(f"The directory {tpm_cert_store} does not exist.")

    for fname in os.listdir(tpm_cert_store):
        if fname.endswith(".pem"):
            break
    else:
        logger.error("The directory %s does not contain any .pem files.", tpm_cert_store)
        raise Exception(f"The directory {tpm_cert_store} does not contain " f"any .pem files")


def cert_loader(tpm_cert_store=config.get("tenant", "tpm_cert_store")):
    file_list = glob.glob(os.path.join(tpm_cert_store, "*.pem"))
    my_trusted_certs = {}
    for file_path in file_list:
        with open(file_path, encoding="utf-8") as f_input:
            my_trusted_certs[file_path] = f_input.read()
    return my_trusted_certs



#!/usr/bin/env python3

import argparse
import configparser
import importlib.util
import json
import os
import re
import sys
from functools import cmp_to_key

from jinja2 import Template

"""
 This script parses the content of a configuration file and use the data to
 replace the values in templates to generate new configuration files.  The
 process is controlled by the "mapping" dictionary, provided through a JSON
 file.

 This dictionary maps the name of the new option to a dictionary with the info
 from the old configuration file format.

 The dictionary has the following fields:
 * "version": The mapping version. Should match the target configuration version
 number
 * "components": A dictionary which keys are the components for which the
 configuration files should be generated. The keys in the "components"
 dictionary should match the template file names without the extension (e.g.
 for the agent component, the key should be "agent" and the template file name
 should be "agent.j2").

 For each component in the "components" dictionary, the value is a dictionary
 which keys are the option names in the new configuration files and the value is
 a dictionary with the information used to find the data in the old
 configuration file. For each option name the value is a dictionary with the
 following fields:

 * "section": The section from the old configuration file format
 * "option": The name of the option in the old configuration file format
 * "default" The default value to use in case the option is missing in the
 provided configuration file

 All values are treated as strings.

 An example could be:

 {
     "version": "1.0",
     "components": {
         "a_component": {
             "some_option": {
                 "section": "some_section",
                 "option" : "some_old_option_name",
                 "default": "default",
             },
             another_option: {
                 "section": "another_section",
                 "option" : "another_old_option_name",
                 "default": "default_value",
             }
         },
         "another_component": {
             "an_option": {
                 "section": "some_section",
                 "option" : "some_old_option_name",
                 "default": "default",
             },
             "another_option": {
                 "section": "another_section",
                 "option" : "another_old_option_name",
                 "default": "default_value",
             }
         }
     }
 }

 Check "scripts/templates/2.0/mapping.json" file for an example.

 The idea is to provide new templates, mapping, and adjust script for new
 versions of the configuration, allowing the user to easily convert from a
 version of the configuration file to the next.
"""

COMPONENTS = ["agent", "verifier", "tenant", "registrar", "ca", "logging"]

# Config files can be merged together, reading from the system to the
# user.
CONFIG_FILES = ["/usr/etc/keylime.conf", "/etc/keylime.conf", os.path.expanduser("~/.config/keylime.conf")]

if "KEYLIME_CONFIG" in os.environ:
    CONFIG_FILES.insert(0, os.environ["KEYLIME_CONFIG"])


def get_config(config_files):
    """
    Read configuration files and merge them together
    """

    flat = [s for ss in config_files for s in ss]
    if not flat:
        print(f"No input provided: using {CONFIG_FILES} as input")
        config_files = CONFIG_FILES
    else:
        config_files = list(set(flat))

    if not any(os.path.exists(c) for c in config_files):
        print(f"Could not find any file of the files in {config_files}. " "Please set input file using --input option.")
        return None
    else:
        config = configparser.RawConfigParser()
        # Validate that at least one config file is present
        config_files = config.read(config_files)
        print(f"Successfully read configuration from {config_files}")

        return config


def output_component(component, config, template, output):
    """
    Output the configuration file for a component
    """

    print(f"Writing {component} configuration to {output}")

    with open(template, "r") as tf:
        t = tf.read()

        j2 = Template(t)

        r = j2.render(config)

        with open(output, "w") as o:
            print(r, file=o)


def output(components, config, templates, outdir):
    """
    Output the requested files
    """

    flat = [s for ss in components for s in ss]

    if not flat:
        # If no component was provided, use all available
        components = COMPONENTS
    else:
        clean = set(flat)
        for c in clean:
            if not c in COMPONENTS:
                print(f"Unknown component {c}, skipping")
        components = list(clean.intersection(COMPONENTS))

    version = config["version"]
    version_dir = os.path.join(templates, version)
    if not os.path.isdir(version_dir):
        raise Exception(f"Could not find directory {version_dir}")

    # Check that there are templates for all components
    for component in components:
        t = os.path.join(version_dir, f"{component}.j2")
        if not os.path.exists(t):
            raise Exception(f"Template file {t} not found")

        # Set output path
        o = os.path.join(outdir, f"{component}.conf")

        output_component(component, config, t, o)


def process_mapping(old_config, templates, mapping_file, debug=False):
    """
    Apply the transformations from the provided mapping file to the
    configuration dictionary
    """

    with open(mapping_file, "r") as f:
        try:
            mapping = json.loads(f.read())
        except Exception as e:
            raise Exception(f"Could not load mapping file {mapping_file}: {e}")

    if not mapping["version"]:
        raise Exception(f'Malformed mapping: no "version" set')

    if not mapping["components"]:
        raise Exception(f'Malformed mapping: no "components" set')

    print(f"Applying mapping from file {mapping_file} version {mapping['version']}")

    # Search for the directory containing the templates for the version set in
    # the mapping
    version_dir = os.path.join(templates, mapping["version"])

    if not os.path.isdir(version_dir):
        raise Exception(
            f"Could not find directory {version_dir} for version " f"{mapping['version']} set in {mapping_file}"
        )

    new = {}

    # Set the new configuration version
    new["version"] = mapping["version"]

    for component in mapping["components"].keys():
        # Create a new dictionary for each component
        new[component] = {}
        m = mapping["components"][component]

        for option in m.keys():
            # For each option, get the dictionary with the info to search:
            # {
            #     "section": section to search
            #     "option": option name
            #     "default": value to use in case the option is missing
            info = m[option]
            try:
                new[component][option] = old_config[info["section"]][info["option"]]
            except Exception as e:
                print(f"[{component}] {e} not found: Using default value \"{info['default']}\" for \"{option}\"")
                new[component][option] = info["default"]

    # If there is an adjust script, load and run it
    adjust_script = os.path.abspath(os.path.join(version_dir, "adjust.py"))
    if os.path.isfile(adjust_script):
        try:
            print(f"Applying adjustment from {adjust_script}")

            # Dynamically load adjust script
            spec = importlib.util.spec_from_file_location("adjust", adjust_script)
            if not spec or not spec.loader:
                raise Exception(f"Could not create spec to load {adjust_script}")

            module = importlib.util.module_from_spec(spec)
            if not module:
                raise Exception(f"Could not load script from {adjust_script}")

            spec.loader.exec_module(module)

            # Run adjust function from adjust script
            execute = getattr(module, "adjust")
            execute(new, mapping)
        except Exception as e:
            print(f"Failed while running adjustment from {adjust_script}: {e}")

    if debug:
        print(json.dumps(new, indent=4))

    return new


def str_to_version(v_str):
    """
    Validates the string format and converts the provided string to a tuple of
    ints which can be sorted and compared.

    :returns: Tuple with version number parts converted to int. In case of
    invalid version string, returns None
    """

    m = re.match(r"^(\d+)\.(\d+)$", v_str)

    if not m:
        return None

    return tuple(int(x) for x in m.group(1, 2))


def process_versions(templates, old_config, target_version=None, debug=False):
    """
    Apply the transformations from the mappings for each version found in the
    templates folder to the configuration.

    If a target version is provided, the process will stop if the target version
    is reached, otherwise all transformations to the latest version are applied
    """

    dirs = os.listdir(templates)

    if not dirs:
        raise Exception(f"No directories found in {templates} for versions")

    # Get a sorted list of the available versions as tuples
    versions = sorted(x for x in set(map(str_to_version, dirs)) if x is not None)

    target = None
    if target_version:
        target = str_to_version(target_version)
        if not target:
            raise Exception(f"Invalid target version number provided: " f"{target_version}")

        if not target in versions:
            raise Exception(f"Directory for target version {target_version} not " f"found in {templates}")

    new = {}

    # If the configuration file does not contain a version number, assume it is
    # in the old format and use the minimum possible version
    try:
        old_version = str_to_version(old_config.get("version", "1.0"))

        if not old_version:
            raise Exception("Invalid version number found in old configuration")

    except (configparser.NoOptionError, configparser.NoSectionError):
        print("No version found in old configuration, using '1.0'")
        old_version = (1, 0)

    for version in versions:
        # Skip versions lower than the current version
        if old_version >= version:
            continue

        # Stop if the version reached the target
        if target:
            if old_version >= target:
                break

        # Find the mapping file for the version and apply
        p = os.path.join(templates, f"{version[0]}.{version[1]}")
        if os.path.isdir(p):
            m = os.path.join(p, "mapping.json")
            if os.path.isfile(m):
                # Apply transformation for the mapping
                new = process_mapping(old_config, templates, m, debug)
                old_config = new
                old_version = str_to_version(new["version"])
                if not old_version:
                    raise Exception(f"Invalid version number '{new['version']}'" f"in {m}")
            else:
                raise Exception(f"Could not find mapping {m}")
        else:
            raise Exception(f"Could not find directory {p}")

    return new


def main():
    parser = argparse.ArgumentParser(description="Split keylime configuration" "file into individual files")

    parser.add_argument(
        "--component",
        help="Select the components for which "
        "configuration files should be generated. If not "
        "provided, generate configuration files for all "
        "components. Can be provided multiple times",
        default=[],
        action="append",
        nargs="+",
    )

    parser.add_argument(
        "--debug",
        help="Print resulting config after each " "applied mapping in JSON format",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--input",
        help="Input keylime configuration file to "
        "process. If not provided, it tries to use the "
        "installed Keylime configuration. Can be provided "
        "multiple times",
        default=[],
        action="append",
        nargs="+",
    )

    parser.add_argument(
        "--out", help="Output directory where to put the " "generated files. Default is current directory", default="."
    )

    parser.add_argument(
        "--mapping",
        help="If provided, then only the provided "
        "mapping file will be used. Otherwise, mappings for "
        'each version found in the "templates" directory will '
        "be applied. The file must be a JSON file containing "
        "the mapping for the new configuration options names "
        "to dictionaries containing the section and option "
        "name from the old configuration file, and the default "
        "value to use in case the option is missing in the "
        "input configuration.",
        default=None,
    )

    parser.add_argument(
        "--templates",
        help="Path to the directory containing "
        "the templates for the configuration files. If not "
        'provided "./templates" is used',
        default="./templates",
    )

    parser.add_argument("--version", help="Target version for the output " "configuration files", default=None)

    args = parser.parse_args()

    if not os.path.exists(args.out):
        raise Exception(f"Output directory {args.out} does not exist")

    if not os.path.isdir(args.out):
        raise Exception(f"File {args.out} is not a directory")

    if args.version and not str_to_version(args.version):
        raise Exception(f"Invalid version {args.version} specified in --version: " f"Expected 'MAJOR.MINOR' format")

    if not os.path.exists(args.templates):
        raise Exception(f"Templates directory {args.templates} does not exist")

    if not os.path.isdir(args.templates):
        raise Exception(f"File {args.templates} is not a directory")

    print(f"Using templates from directory {args.templates}")

    # Get old configuration
    old_config = get_config(args.input)
    if not old_config:
        # None of the provided files were parsed successfully
        return

    if args.mapping:
        if os.path.isfile(args.mapping):
            mapping_file = args.mapping
            # Apply the single mapping provided
            config = process_mapping(old_config, args.templates, mapping_file, debug=args.debug)
        else:
            raise Exception(f"Could not find provided mapping {args.mapping}")
    else:
        # Apply transformations from the templates in the templates directory
        # If a target version is provided, stop when reaching the target
        config = process_versions(args.templates, old_config, target_version=args.version, debug=args.debug)

    output(args.component, config, args.templates, args.out)


if __name__ == "__main__":
    main()



import base64
import binascii
import codecs
import collections
import hashlib
import os
import re
import sys
import tempfile
import threading
import time
import typing
import zlib

from cryptography.hazmat.primitives import serialization as crypto_serialization
from packaging.version import Version

from keylime import cert_utils, cmd_exec, config, keylime_logging, secure_mount
from keylime.agentstates import TPMClockInfo
from keylime.common import algorithms, retry
from keylime.failure import Component, Failure
from keylime.tpm import tpm2_objects, tpm_abstract

logger = keylime_logging.init_logging("tpm")


def _get_cmd_env():
    env = os.environ.copy()
    if "TPM2TOOLS_TCTI" not in env:
        # Don't clobber existing setting (if present)
        env["TPM2TOOLS_TCTI"] = "device:/dev/tpmrm0"
        # env['TPM2TOOLS_TCTI'] = 'tabrmd:bus_name=com.intel.tss2.Tabrmd'
        # Other (not recommended) options are direct emulator and chardev communications:
        # env['TPM2TOOLS_TCTI'] = 'mssim:port=2321'
        # env['TPM2TOOLS_TCTI'] = 'device:/dev/tpm0'
    return env


class tpm(tpm_abstract.AbstractTPM):
    VERSION = 2
    tools_version = ""

    def __init__(self, need_hw_tpm=False):
        tpm_abstract.AbstractTPM.__init__(self, need_hw_tpm)

        # Shared lock to serialize access to tools
        self.tpmutilLock = threading.Lock()

        self.__get_tpm2_tools()

    def __get_tpm2_tools(self):
        retDict = self.__run(["tpm2_startup", "--version"])

        code = retDict["code"]
        output = "".join(config.convert(retDict["retout"]))
        errout = "".join(config.convert(retDict["reterr"]))
        if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
            raise Exception(
                "Error establishing tpm2-tools version using TPM2_Startup: %s" + str(code) + ": " + str(errout)
            )

        # Extract the `version="x.x.x"` from tools
        version_str = re.search(r'version="([^"]+)"', output).group(1)
        # Extract the full semver release number.
        self.tools_version = version_str.split("-")

        if Version(self.tools_version[0]) >= Version("5.4") or (
            # Also mark first git version that introduces the change to the tpm2_eventlog format as 5.4
            # See: https://github.com/tpm2-software/tpm2-tools/commit/c78d258b2588aee535fd17594ad2f5e808056373
            Version(self.tools_version[0]) == Version("5.3")
            and len(self.tools_version) > 1
            and int(self.tools_version[1]) >= 24
        ):
            logger.info("TPM2-TOOLS Version: %s", self.tools_version[0])
            self.tools_version = "5.4"
        elif Version(self.tools_version[0]) >= Version("4.2"):
            logger.info("TPM2-TOOLS Version: %s", self.tools_version[0])
            self.tools_version = "4.2"
        elif Version(self.tools_version[0]) >= Version("4.0.0"):
            logger.info("TPM2-TOOLS Version: %s", self.tools_version[0])
            self.tools_version = "4.0"
        elif Version(self.tools_version[0]) >= Version("3.2.0"):
            logger.info("TPM2-TOOLS Version: %s", self.tools_version[0])
            self.tools_version = "3.2"
        else:
            logger.error("TPM2-TOOLS Version %s is not supported.", self.tools_version[0])
            sys.exit()

    def __get_tpm_algorithms(self):
        if self.tools_version == "3.2":
            retDict = self.__run(["tpm2_getcap", "-c", "algorithms"])
        elif self.tools_version in ["4.0", "4.2", "5.4"]:
            retDict = self.__run(["tpm2_getcap", "algorithms"])

        output = config.convert(retDict["retout"])
        errout = config.convert(retDict["reterr"])
        code = retDict["code"]

        if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
            raise Exception("get_tpm_algorithms failed with code " + str(code) + ": " + str(errout))

        if self.tools_version == "3.2":
            # output, human-readable -> json
            output = "".join(output)
            output = re.sub(r"TPMA_ALGORITHM for ALG_ID: 0x[0-9a-f]+\s+-\s+([a-z0-9_]+)", r"\1:", output)
            output = output.replace("set", "1")
            output = output.replace("clear", "0")
            output = [output]

        retyaml = config.yaml_to_dict(output, logger=logger)
        if retyaml is None:
            logger.warning("Could not read YAML output of tpm2_getcap.")
            return
        for algorithm, details in retyaml.items():
            if details["asymmetric"] == 1 and details["object"] == 1 and algorithms.Encrypt.is_recognized(algorithm):
                self.supported["encrypt"].add(algorithm)
            elif details["hash"] == 1 and algorithms.Hash.is_recognized(algorithm):
                self.supported["hash"].add(algorithm)
            elif details["asymmetric"] == 1 and details["signing"] == 1 and algorithms.Sign.is_recognized(algorithm):
                self.supported["sign"].add(algorithm)

    def __get_pcrs(self):
        """Gets which PCRs are enabled with which hash algorithm"""
        if self.tools_version == "3.2":
            retDict = self.__run(["tpm2_getcap", "-c", "pcrs"])
        elif self.tools_version in ["4.0", "4.2", "5.4"]:
            retDict = self.__run(["tpm2_getcap", "pcrs"])

        output = config.convert(retDict["retout"])
        errout = config.convert(retDict["reterr"])
        code = retDict["code"]

        if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
            raise Exception("get_tpm_algorithms failed with code " + str(code) + ": " + str(errout))

        retyaml = config.yaml_to_dict(output, logger=logger)
        pcrs = {}
        if retyaml is None:
            logger.warning("Could not read YAML output of tpm2_getcap.")
            return pcrs
        if "selected-pcrs" in retyaml:
            pcrs = collections.ChainMap(*retyaml["selected-pcrs"])
        return pcrs

    def run(self, cmd):
        return self.__run(cmd, lock=False)

    def __run(
        self, cmd, expectedcode=tpm_abstract.AbstractTPM.EXIT_SUCESS, raiseOnError=True, lock=True, outputpaths=None
    ):
        env = _get_cmd_env()

        # Convert single outputpath to list
        if isinstance(outputpaths, str):
            outputpaths = [outputpaths]

        numtries = 0
        while True:
            if lock:
                with self.tpmutilLock:
                    retDict = cmd_exec.run(
                        cmd=cmd, expectedcode=expectedcode, raiseOnError=False, outputpaths=outputpaths, env=env
                    )
            else:
                retDict = cmd_exec.run(
                    cmd=cmd, expectedcode=expectedcode, raiseOnError=False, outputpaths=outputpaths, env=env
                )
            code = retDict["code"]
            retout = retDict["retout"]
            reterr = retDict["reterr"]

            # keep trying to get quote if a PCR race condition occurred in quote
            if cmd[0] == "tpm2_quote" and cmd_exec.list_contains_substring(
                reterr, "Error validating calculated PCR composite with quote"
            ):
                numtries += 1
                maxr = config.getint("agent", "max_retries")
                if numtries >= maxr:
                    logger.error("Agent did not return proper quote due to PCR race condition.")
                    break
                interval = config.getfloat("agent", "retry_interval")
                exponential_backoff = config.getboolean("agent", "exponential_backoff")
                next_retry = retry.retry_time(exponential_backoff, interval, numtries, logger)
                logger.info(
                    "Failed to get quote %d/%d times, trying again in %f seconds...", numtries, maxr, next_retry
                )
                time.sleep(next_retry)
                continue

            break

        # Don't bother continuing if TPM call failed and we're raising on error
        if code != expectedcode and raiseOnError:
            raise Exception(
                f"Command: {cmd} returned {code}, expected {expectedcode}, output {retout}, stderr {reterr}"
            )

        return retDict

    # tpm_initialize
    def __startup_tpm(self):
        retDict = self.__run(["tpm2_startup", "-c"])
        errout = config.convert(retDict["reterr"])
        code = retDict["code"]
        if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
            raise Exception("Error initializing emulated TPM with TPM2_Startup: %s" + str(code) + ": " + str(errout))

    def __create_ek(self, asym_alg=None):
        # this function is intended to be idempotent
        if asym_alg is None:
            asym_alg = self.defaults["encrypt"]

        current_handle = self.get_tpm_metadata("ek_handle")
        owner_pw = self.get_tpm_metadata("owner_pw")

        # clear out old handle before starting again (give idempotence)
        if current_handle is not None and owner_pw is not None:
            logger.info("Flushing old ek handle: %s", hex(current_handle))
            if self.tools_version == "3.2":
                retDict = self.__run(["tpm2_getcap", "-c", "handles-persistent"], raiseOnError=False)
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                retDict = self.__run(["tpm2_getcap", "handles-persistent"], raiseOnError=False)
            output = retDict["retout"]
            reterr = retDict["reterr"]
            code = retDict["code"]

            if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                raise Exception("tpm2_getcap failed with code " + str(code) + ": " + str(reterr))

            outjson = config.yaml_to_dict(output, logger=logger)
            if outjson is not None and hex(current_handle) in outjson:
                if self.tools_version == "3.2":
                    cmd = ["tpm2_evictcontrol", "-A", "o", "-H", hex(current_handle), "-P", owner_pw]
                    retDict = self.__run(cmd, raiseOnError=False)
                elif self.tools_version in ["4.0", "4.2", "5.4"]:
                    cmd = ["tpm2_evictcontrol", "-C", "o", "-c", hex(current_handle), "-P", owner_pw]
                    retDict = self.__run(cmd, raiseOnError=False)
                output = retDict["retout"]
                reterr = retDict["reterr"]
                code = retDict["code"]

                if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                    logger.info(
                        "Failed to flush old ek handle: %s.  Code %s: %s", hex(current_handle), str(code), str(reterr)
                    )

                self._set_tpm_metadata("ek_handle", None)
                self._set_tpm_metadata("ek_tpm", None)
                self._set_tpm_metadata("ek_pw", None)

        # make sure an ownership pw is set
        if owner_pw is None:
            owner_pw = tpm_abstract.TPM_Utilities.random_password(20)
            self._set_tpm_metadata("owner_pw", owner_pw)
        ek_pw = tpm_abstract.TPM_Utilities.random_password(20)

        # create a new ek
        with tempfile.NamedTemporaryFile() as tmppath:
            # TODO(kaifeng) Missing else here for other versions
            if self.tools_version == "3.2":
                command = [
                    "tpm2_getpubek",
                    "-H",
                    "0x81010007",
                    "-g",
                    asym_alg,
                    "-f",
                    tmppath.name,
                    "-P",
                    ek_pw,
                    "-o",
                    owner_pw,
                    "-e",
                    owner_pw,
                ]
            elif self.tools_version == "4.0":
                command = [
                    "tpm2_createek",
                    "-c",
                    "-",
                    "-G",
                    asym_alg,
                    "-u",
                    tmppath.name,
                    "-p",
                    ek_pw,
                    "-w",
                    owner_pw,
                    "-P",
                    owner_pw,
                ]
            elif self.tools_version == "4.2":
                command = [
                    "tpm2_createek",
                    "-c",
                    "-",
                    "-G",
                    asym_alg,
                    "-u",
                    tmppath.name,
                    "-w",
                    owner_pw,
                    "-P",
                    owner_pw,
                ]

            retDict = self.__run(command, raiseOnError=False, outputpaths=tmppath.name)
            output = retDict["retout"]
            reterr = retDict["reterr"]
            code = retDict["code"]
            ek_tpm = retDict["fileouts"][tmppath.name]

            if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                raise Exception("createek failed with code " + str(code) + ": " + str(reterr))

            if self.tools_version == "3.2":
                handle = int(0x81010007)
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                handle = None
                retyaml = config.yaml_to_dict(output, logger=logger)
                if retyaml is None:
                    raise Exception("Could not read YAML output of tpm2_createek.")
                if "persistent-handle" in retyaml:
                    handle = retyaml["persistent-handle"]

            # Make sure that all transient objects are flushed
            self.__run(["tpm2_flushcontext", "-t"], raiseOnError=False)

            self._set_tpm_metadata("ek_handle", handle)
            self._set_tpm_metadata("ek_pw", ek_pw)
            self._set_tpm_metadata("ek_tpm", base64.b64encode(ek_tpm))
            self._set_tpm_metadata("ek_alg", asym_alg)

    def __use_ek(self, ek_handle, config_pw):
        ek_handle = int(ek_handle, 16)
        logger.info("Using an already created ek with handle: %s", hex(ek_handle))

        self._set_tpm_metadata("owner_pw", config_pw)

        with tempfile.NamedTemporaryFile() as tmppath:
            if self.tools_version == "3.2":
                cmd = ["tpm2_readpublic", "-H", hex(ek_handle), "-o", tmppath.name, "-f", "tss"]
                retDict = self.__run(cmd, raiseOnError=False, outputpaths=tmppath.name)
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                cmd = ["tpm2_readpublic", "-c", hex(ek_handle), "-o", tmppath.name, "-f", "tss"]
                retDict = self.__run(cmd, raiseOnError=False, outputpaths=tmppath.name)

            reterr = retDict["reterr"]
            code = retDict["code"]
            ek_tpm = retDict["fileouts"][tmppath.name]
            if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                raise Exception("tpm2_readpublic failed with code " + str(code) + ": " + str(reterr))
            self._set_tpm_metadata("ek_tpm", base64.b64encode(ek_tpm))

        self._set_tpm_metadata("ek_handle", int(ek_handle))

    def __take_ownership(self, config_pw):
        # if no ownerpassword
        if config_pw == "generate":
            logger.info("Generating random TPM owner password")
            owner_pw = tpm_abstract.TPM_Utilities.random_password(20)
        else:
            logger.info("Taking ownership with config provided TPM owner password")
            owner_pw = config_pw

        logger.debug("Removing all saved sessions from TPM")
        retDict = self.__run(["tpm2_flushcontext", "-s"], raiseOnError=False)

        if self.tools_version == "3.2":
            retDict = self.__run(["tpm2_takeownership", "-c"], raiseOnError=False)
            retDict = self.__run(["tpm2_takeownership", "-o", owner_pw, "-e", owner_pw], raiseOnError=False)
        elif self.tools_version in ["4.0", "4.2", "5.4"]:
            retDict = self.__run(["tpm2_changeauth", "-c", "o", owner_pw], raiseOnError=False)
            retDict = self.__run(["tpm2_changeauth", "-c", "e", owner_pw], raiseOnError=False)

        code = retDict["code"]
        if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
            # if we fail, see if already owned with this pw
            if self.tools_version == "3.2":
                retDict = self.__run(
                    ["tpm2_takeownership", "-o", owner_pw, "-e", owner_pw, "-O", owner_pw, "-E", owner_pw],
                    raiseOnError=False,
                )
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                retDict = self.__run(["tpm2_changeauth", "-c", "o", "-p", owner_pw, owner_pw], raiseOnError=False)
                retDict = self.__run(["tpm2_changeauth", "-c", "e", "-p", owner_pw, owner_pw], raiseOnError=False)

            reterr = retDict["reterr"]
            code = retDict["code"]
            if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                # ut-oh, already owned but not with provided pw!
                raise Exception("Owner password unknown, TPM reset required. Code %s" + str(code) + ": " + str(reterr))

        self._set_tpm_metadata("owner_pw", owner_pw)
        logger.info("TPM Owner password confirmed: %s", owner_pw)

    def __get_pub_ek(self):  # assumes that owner_pw is correct at this point
        handle = self.get_tpm_metadata("ek_handle")
        if handle is None:
            raise Exception("create_ek has not been run yet?")
        # make a temp file for the output
        with tempfile.NamedTemporaryFile() as tmppath:
            # generates pubek.pem
            if self.tools_version == "3.2":
                cmd = ["tpm2_readpublic", "-H", hex(handle), "-o", tmppath.name]
                retDict = self.__run(cmd, raiseOnError=False, outputpaths=tmppath.name)
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                cmd = ["tpm2_readpublic", "-c", hex(handle), "-o", tmppath.name]
                retDict = self.__run(cmd, raiseOnError=False, outputpaths=tmppath.name)

            reterr = retDict["reterr"]
            code = retDict["code"]
            ek = retDict["fileouts"][tmppath.name]
            if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                raise Exception("tpm2_readpublic failed with code " + str(code) + ": " + str(reterr))

        self._set_tpm_metadata("ek_tpm", base64.b64encode(ek))

    def __create_aik(self, asym_alg=None, hash_alg=None, sign_alg=None):
        if hash_alg is None:
            hash_alg = self.defaults["hash"]
        if asym_alg is None:
            asym_alg = self.defaults["encrypt"]
        if sign_alg is None:
            sign_alg = self.defaults["sign"]

        owner_pw = self.get_tpm_metadata("owner_pw")

        # clear out old handle before starting again (give idempotence)
        if self.get_tpm_metadata("aik_handle") is not None:
            aik_handle = self.get_tpm_metadata("aik_handle")
            if self.tools_version == "3.2":
                logger.info("Flushing old ak handle: %s", hex(aik_handle))
                retDict = self.__run(["tpm2_getcap", "-c", "handles-persistent"], raiseOnError=False)
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                logger.info("Flushing old ak handle: %s", aik_handle)
                retDict = self.__run(["tpm2_getcap", "handles-persistent"], raiseOnError=False)
            output = config.convert(retDict["retout"])
            errout = config.convert(retDict["reterr"])
            code = retDict["code"]

            if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                raise Exception("tpm2_getcap failed with code " + str(code) + ": " + str(errout))

            if self.tools_version == "3.2":
                # output, human-readable -> json
                output = "".join(output)
                output = output.replace("0x", " - 0x")
                output = [output]

            outjson = config.yaml_to_dict(output, logger=logger)
            if self.tools_version == "3.2":
                evict_it = outjson is not None and aik_handle in outjson
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                evict_it = os.path.exists(aik_handle)
            if evict_it:
                if self.tools_version == "3.2":
                    cmd = ["tpm2_evictcontrol", "-A", "o", "-H", hex(aik_handle), "-P", owner_pw]
                    retDict = self.__run(cmd, raiseOnError=False)
                elif self.tools_version in ["4.0", "4.2", "5.4"]:
                    cmd = ["tpm2_evictcontrol", "-C", "o", "-c", aik_handle, "-P", owner_pw]
                    retDict = self.__run(cmd, raiseOnError=False)
                    os.remove(aik_handle)

                output = retDict["retout"]
                reterr = retDict["reterr"]
                code = retDict["code"]

                if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                    if self.tools_version == "3.2":
                        logger.info(
                            "Failed to flush old ak handle: %s.  Code %s: %s", hex(aik_handle), str(code), str(reterr)
                        )
                    elif self.tools_version in ["4.0", "4.2", "5.4"]:
                        logger.info(
                            "Failed to flush old ak handle: %s.  Code %s: %s", aik_handle, str(code), str(reterr)
                        )

                self._set_tpm_metadata("aik_pw", None)
                self._set_tpm_metadata("aik_tpm", None)
                self._set_tpm_metadata("aik_handle", None)

        logger.debug("Creating a new AIK identity")

        # We need an ek handle to make an aik
        ek_handle = self.get_tpm_metadata("ek_handle")
        if ek_handle is None:
            raise Exception("Failed to create AIK, since EK has not yet been created!")

        aik_pw = tpm_abstract.TPM_Utilities.random_password(20)
        # make a temp file for the output
        with tempfile.NamedTemporaryFile() as akpubfile:
            secpath = ""
            if self.tools_version in ["4.0", "4.2", "5.4"]:
                # ok lets write out the key now
                secdir = secure_mount.mount()  # confirm that storage is still securely mounted
                secfd, secpath = tempfile.mkstemp(dir=secdir)

            if self.tools_version == "3.2":
                command = [
                    "tpm2_getpubak",
                    "-E",
                    hex(ek_handle),
                    "-k",
                    "0x81010008",
                    "-g",
                    asym_alg,
                    "-D",
                    hash_alg,
                    "-s",
                    sign_alg,
                    "-f",
                    akpubfile.name,
                    "-e",
                    owner_pw,
                    "-P",
                    aik_pw,
                    "-o",
                    owner_pw,
                ]
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                command = [
                    "tpm2_createak",
                    "-C",
                    hex(ek_handle),
                    "-c",
                    secpath,
                    "-G",
                    asym_alg,
                    "-g",
                    hash_alg,
                    "-s",
                    sign_alg,
                    "-u",
                    akpubfile.name,
                    "-p",
                    aik_pw,
                    "-P",
                    owner_pw,
                ]
            retDict = self.__run(command, outputpaths=akpubfile.name)
            if secfd >= 0:
                os.close(secfd)
            retout = retDict["retout"]
            reterr = retDict["reterr"]
            code = retDict["code"]

            if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                raise Exception("tpm2_createak failed with code " + str(code) + ": " + str(reterr))

            jsonout = config.yaml_to_dict(retout, logger=logger)
            if jsonout is None:
                raise Exception(
                    "unable to parse YAML output of tpm2_createak. Is your tpm2-tools installation up to date?"
                )
            aik_tpm = retDict["fileouts"][akpubfile.name]
            if aik_tpm == "":
                raise Exception(
                    "unable to read public aik from create identity.  Is your tpm2-tools installation up to date?"
                )
            self._set_tpm_metadata("aik_tpm", base64.b64encode(aik_tpm))

        if self.tools_version == "3.2":
            if "loaded-key" not in jsonout or "name" not in jsonout["loaded-key"]:
                raise Exception("tpm2_createak failed to create aik: return " + str(reterr))

            handle = int(0x81010008)

            # get and persist the pem (not returned by tpm2_getpubak)
            self._set_tpm_metadata("aik_handle", handle)
        elif self.tools_version in ["4.0", "4.2", "5.4"]:
            if "loaded-key" not in jsonout:
                raise Exception("tpm2_createak failed to create aik: return " + str(reterr))

            handle = secpath

            # persist the pem
            self._set_tpm_metadata("aik_handle", handle)

        # Make sure that all transient objects are flushed
        self.__run(["tpm2_flushcontext", "-t"], raiseOnError=False)

        # persist common results
        self._set_tpm_metadata("aik_pw", aik_pw)

    def flush_keys(self):
        logger.debug("Flushing keys from TPM...")
        if self.tools_version == "3.2":
            retDict = self.__run(["tpm2_getcap", "-c", "handles-persistent"])
        elif self.tools_version in ["4.0", "4.2", "5.4"]:
            retDict = self.__run(["tpm2_getcap", "handles-persistent"])
        # retout = retDict['retout']
        retout = config.convert(retDict["retout"])
        errout = config.convert(retDict["reterr"])
        code = retDict["code"]

        if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
            logger.debug("tpm2_getcap failed with code %s: %s", str(code), str(errout))

        if self.tools_version == "3.2":
            # output, human-readable -> json
            retout = "".join(retout)
            retout = retout.replace("0x", " - 0x")
            retout = [retout]

        owner_pw = self.get_tpm_metadata("owner_pw")
        jsonout = config.yaml_to_dict(retout, logger=logger)
        if jsonout is None:
            logger.warning("Could not read YAML output of tpm2_getcap.")
            jsonout = {}
        for key in jsonout:
            if str(hex(key)) != self.defaults["ek_handle"]:
                logger.debug("Flushing key handle %s", hex(key))
                if self.tools_version == "3.2":
                    self.__run(["tpm2_evictcontrol", "-A", "o", "-H", hex(key), "-P", owner_pw], raiseOnError=False)
                elif self.tools_version in ["4.0", "4.2", "5.4"]:
                    self.__run(["tpm2_evictcontrol", "-C", "o", "-c", hex(key), "-P", owner_pw], raiseOnError=False)
        # Make sure that all transient objects are flushed
        self.__run(["tpm2_flushcontext", "-t"], lock=False, raiseOnError=False)

    def encryptAIK(self, uuid, ek_tpm: bytes, aik_tpm: bytes):

        if ek_tpm is None or aik_tpm is None:
            logger.error("Missing parameters for encryptAIK")
            return None

        aik_name = tpm2_objects.get_tpm2b_public_name(aik_tpm)

        efd = keyfd = blobfd = -1
        ekFile = None
        challengeFile = None
        keyblob = None
        blobpath = None

        try:
            # write out the public EK
            efd, etemp = tempfile.mkstemp()
            with open(etemp, "wb") as ekFile:
                ekFile.write(ek_tpm)

            # write out the challenge
            challenge = tpm_abstract.TPM_Utilities.random_password(32)
            challenge = challenge.encode()
            keyfd, keypath = tempfile.mkstemp()
            with open(keypath, "wb") as challengeFile:
                challengeFile.write(challenge)

            # create temp file for the blob
            blobfd, blobpath = tempfile.mkstemp()
            command = [
                "tpm2_makecredential",
                "-T",
                "none",
                "-e",
                ekFile.name,
                "-s",
                challengeFile.name,
                "-n",
                aik_name,
                "-o",
                blobpath,
            ]
            self.__run(command, lock=False)

            logger.info("Encrypting AIK for UUID %s", uuid)

            # read in the blob
            with open(blobpath, "rb") as f:
                keyblob = base64.b64encode(f.read())

            # read in the aes key
            key = base64.b64encode(challenge).decode("utf-8")

        except Exception as e:
            logger.error("Error encrypting AIK: %s", str(e))
            logger.exception(e)
            raise
        finally:
            for fd in [efd, keyfd, blobfd]:
                if fd >= 0:
                    os.close(fd)
            for fi in [ekFile, challengeFile]:
                if fi is not None:
                    os.remove(fi.name)
            if blobpath is not None:
                os.remove(blobpath)

        return (keyblob, key)

    def activate_identity(self, keyblob):
        owner_pw = self.get_tpm_metadata("owner_pw")
        aik_keyhandle = self.get_tpm_metadata("aik_handle")
        ek_keyhandle = self.get_tpm_metadata("ek_handle")

        keyblobFile = None
        secpath = None
        secfd = -1
        sesspath = None
        sesspathfd = -1
        try:
            # write out key blob
            kfd, ktemp = tempfile.mkstemp()
            with open(ktemp, "wb") as keyblobFile:
                # the below is a coroutine?
                keyblobFile.write(base64.b64decode(keyblob))

            os.close(kfd)

            # ok lets write out the key now
            secdir = secure_mount.mount()  # confirm that storage is still securely mounted

            secfd, secpath = tempfile.mkstemp(dir=secdir)
            sesspathfd, sesspath = tempfile.mkstemp(dir=secdir)

            apw = self.get_tpm_metadata("aik_pw")
            if self.tools_version == "3.2":
                command = [
                    "tpm2_activatecredential",
                    "-H",
                    hex(aik_keyhandle),
                    "-k",
                    hex(ek_keyhandle),
                    "-f",
                    keyblobFile.name,
                    "-o",
                    secpath,
                    "-P",
                    apw,
                    "-e",
                    owner_pw,
                ]
                retDict = self.__run(command, outputpaths=secpath)
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                self.__run(["tpm2_startauthsession", "--policy-session", "-S", sesspath])
                self.__run(["tpm2_policysecret", "-S", sesspath, "-c", "0x4000000B", owner_pw])
                command = [
                    "tpm2_activatecredential",
                    "-c",
                    aik_keyhandle,
                    "-C",
                    hex(ek_keyhandle),
                    "-i",
                    keyblobFile.name,
                    "-o",
                    secpath,
                    "-p",
                    apw,
                    "-P",
                    f"session:{sesspath}",
                ]
                retDict = self.__run(command, outputpaths=secpath)
                self.__run(["tpm2_flushcontext", sesspath])

            # Make sure that all transient objects are flushed
            self.__run(["tpm2_flushcontext", "-t"], raiseOnError=False)

            fileout = retDict["fileouts"][secpath]
            logger.info("AIK activated.")

            key = base64.b64encode(fileout)

        except Exception as e:
            logger.error("Error decrypting AIK: %s", str(e))
            logger.exception(e)
            return None
        finally:
            if keyblobFile is not None:
                os.remove(keyblobFile.name)
            if secfd >= 0:
                os.close(secfd)
            if secpath is not None and os.path.exists(secpath):
                os.remove(secpath)
            if sesspathfd >= 0:
                os.close(sesspathfd)
            if sesspath is not None and os.path.exists(sesspath):
                os.remove(sesspath)
        return key

    def verify_ek(self, ekcert):
        """Verify that the provided EK certificate is signed by a trusted root
        :param ekcert: The Endorsement Key certificate in DER format
        :returns: True if the certificate can be verified, false otherwise
        """
        return cert_utils.verify_ek(ekcert)

    def get_tpm_manufacturer(self, output=None):
        vendorStr = None
        if not output:
            if self.tools_version == "3.2":
                retDict = self.__run(["tpm2_getcap", "-c", "properties-fixed"])
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                retDict = self.__run(["tpm2_getcap", "properties-fixed"])
            else:
                raise Exception(f"Unsupported tools version: {self.tools_version}")

            output = retDict["retout"]
            reterr = retDict["reterr"]
            code = retDict["code"]

            if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                raise Exception("get_tpm_manufacturer failed with code " + str(code) + ": " + str(reterr))

        # Clean up TPM manufacturer information (strip control characters)
        # These strings are supposed to be printable ASCII characters, but
        # some TPM manufacturers put control characters in here
        #
        # TPM manufacturer information can also contain un-escaped
        # double quotes. Making sure that un-escaped quotes are
        # replaced before attempting YAML parse.
        def quoterepl(m):
            return '"' + m.group(0)[1:-1].replace('"', '\\"') + '"'

        for i, s in enumerate(output):
            s1 = re.sub(r'(?!".*\\".*")".*".*"', quoterepl, s.decode("utf-8"))
            output[i] = re.sub(r"[\x01-\x1F\x7F]", "", s1).encode("utf-8")

        retyaml = config.yaml_to_dict(output, logger=logger)
        if retyaml is None:
            raise Exception("Could not read YAML output of tpm2_getcap.")
        if "TPM2_PT_VENDOR_STRING_1" in retyaml:
            vendorStr = retyaml["TPM2_PT_VENDOR_STRING_1"]["value"]
        elif "TPM_PT_VENDOR_STRING_1" in retyaml:
            vendorStr = retyaml["TPM_PT_VENDOR_STRING_1"]["as string"].strip()

        return vendorStr

    def is_emulator(self):
        return self.get_tpm_manufacturer() == "SW"

    def tpm_init(self, self_activate=False, config_pw=None):
        # this was called tpm_initialize.init before
        self.warn_emulator()

        self.defaults["ek_handle"] = config.get("agent", "ek_handle")

        if self.need_hw_tpm:
            # We don't know which algs the TPM supports yet
            self.supported["encrypt"] = set()
            self.supported["hash"] = set()
            self.supported["sign"] = set()

            # Grab which default algs the config requested
            defaultHash = config.get("agent", "tpm_hash_alg")
            defaultEncrypt = config.get("agent", "tpm_encryption_alg")
            defaultSign = config.get("agent", "tpm_signing_alg")

            if self.defaults["ek_handle"] == "generate":
                # Start up the TPM
                self.__startup_tpm()

            # Figure out which algorithms the TPM supports
            self.__get_tpm_algorithms()

            # Ensure TPM supports the defaults requested
            if defaultHash not in self.supported["hash"]:
                raise Exception(f"Unsupported hash algorithm specified: {str(defaultHash)}!")

            if defaultEncrypt not in self.supported["encrypt"]:
                raise Exception(f"Unsupported encryption algorithm specified: {str(defaultEncrypt)}!")

            if defaultSign not in self.supported["sign"]:
                raise Exception(f"Unsupported signing algorithm specified: {str(defaultSign)}!")

            enabled_pcrs = self.__get_pcrs()
            if not enabled_pcrs.get(str(defaultHash)):
                raise Exception(f"No PCR banks enabled for hash algorithm specified: {defaultHash}")

            self.defaults["hash"] = algorithms.Hash(defaultHash)
            self.defaults["encrypt"] = defaultEncrypt
            self.defaults["sign"] = defaultSign

        if self.defaults["ek_handle"] == "generate":
            self.__take_ownership(config_pw)
            self.__create_ek()
        else:
            self.__use_ek(self.defaults["ek_handle"], config_pw)

        self.__get_pub_ek()

        ekcert = self.read_ekcert_nvram()
        self._set_tpm_metadata("ekcert", ekcert)

        # if no AIK created, then create one
        self.__create_aik()

        return self.get_tpm_metadata("ekcert"), self.get_tpm_metadata("ek_tpm"), self.get_tpm_metadata("aik_tpm")

    # tpm_quote
    @staticmethod
    def __pcr_mask_to_list(mask):
        pcr_list = []
        for pcr in range(24):
            if tpm_abstract.TPM_Utilities.check_mask(mask, pcr):
                pcr_list.append(str(pcr))
        return ",".join(pcr_list)

    def create_quote(self, nonce, data=None, pcrmask=tpm_abstract.AbstractTPM.EMPTYMASK, hash_alg=None, compress=False):
        if hash_alg is None:
            hash_alg = self.defaults["hash"]

        quote = ""

        with tempfile.NamedTemporaryFile() as quotepath, tempfile.NamedTemporaryFile() as sigpath, tempfile.NamedTemporaryFile() as pcrpath:
            keyhandle = self.get_tpm_metadata("aik_handle")
            aik_pw = self.get_tpm_metadata("aik_pw")

            if pcrmask is None:
                pcrmask = tpm_abstract.AbstractTPM.EMPTYMASK

            if data is not None:
                # add PCR 16 to pcrmask
                pcrmask = hex(int(pcrmask, 0) | (1 << config.TPM_DATA_PCR))

            pcrlist = self.__pcr_mask_to_list(pcrmask)

            with self.tpmutilLock:
                if data is not None:
                    self.__run(["tpm2_pcrreset", str(config.TPM_DATA_PCR)], lock=False)
                    self.extendPCR(pcrval=config.TPM_DATA_PCR, hashval=self.hashdigest(data), lock=False)

                nonce = bytes(nonce, encoding="utf8").hex()
                if self.tools_version == "3.2":
                    command = [
                        "tpm2_quote",
                        "-k",
                        hex(keyhandle),
                        "-L",
                        f"{str(hash_alg)}:{pcrlist}",
                        "-q",
                        nonce,
                        "-m",
                        quotepath.name,
                        "-s",
                        sigpath.name,
                        "-p",
                        pcrpath.name,
                        "-G",
                        hash_alg,
                        "-P",
                        aik_pw,
                    ]
                elif self.tools_version in ["4.0", "4.2", "5.4"]:
                    command = [
                        "tpm2_quote",
                        "-c",
                        keyhandle,
                        "-l",
                        f"{str(hash_alg)}:{pcrlist}",
                        "-q",
                        nonce,
                        "-m",
                        quotepath.name,
                        "-s",
                        sigpath.name,
                        "-o",
                        pcrpath.name,
                        "-g",
                        hash_alg,
                        "-p",
                        aik_pw,
                    ]
                retDict = self.__run(command, lock=False, outputpaths=[quotepath.name, sigpath.name, pcrpath.name])
                # Make sure that all transient objects are flushed
                self.__run(["tpm2_flushcontext", "-t"], lock=False, raiseOnError=False)
                quoteraw = retDict["fileouts"][quotepath.name]
                sigraw = retDict["fileouts"][sigpath.name]
                pcrraw = retDict["fileouts"][pcrpath.name]
                if compress:
                    quoteraw = zlib.compress(quoteraw)
                    sigraw = zlib.compress(sigraw)
                    pcrraw = zlib.compress(pcrraw)
                quote_b64encode = base64.b64encode(quoteraw)
                sigraw_b64encode = base64.b64encode(sigraw)
                pcrraw_b64encode = base64.b64encode(pcrraw)
                quote = (
                    quote_b64encode.decode("utf-8")
                    + ":"
                    + sigraw_b64encode.decode("utf-8")
                    + ":"
                    + pcrraw_b64encode.decode("utf-8")
                )

        return "r" + quote

    def __tpm2_checkquote(self, pubaik, nonce, quoteFile, sigFile, pcrFile, hash_alg):
        nonce = bytes(nonce, encoding="utf8").hex()
        if self.tools_version == "3.2":
            command = [
                "tpm2_checkquote",
                "-c",
                pubaik,
                "-m",
                quoteFile,
                "-s",
                sigFile,
                "-p",
                pcrFile,
                "-G",
                hash_alg,
                "-q",
                nonce,
            ]
        elif self.tools_version in ["4.0", "4.2", "5.4"]:
            command = [
                "tpm2_checkquote",
                "-u",
                pubaik,
                "-m",
                quoteFile,
                "-s",
                sigFile,
                "-f",
                pcrFile,
                "-g",
                hash_alg,
                "-q",
                nonce,
            ]
        retDict = self.__run(command, lock=False)
        return retDict

    def __tpm2_printquote(self, quoteFile):
        command = ["tpm2_print", "-t", "TPMS_ATTEST", quoteFile]
        retDict = self.__run(command, lock=False)
        return retDict

    def _tpm2_printquote(self, quote, compressed):
        """Get TPM timestamp info from quote
        :param quote: quote data in the format 'r<b64-compressed-quoteblob>:<b64-compressed-sigblob>:<b64-compressed-pcrblob>
        :param compressed: if the quote data is compressed with zlib or not
        :returns: Returns the 'retout' from running tpm2_print and True in case of success, None and False in case of error.
        This function throws an Exception on bad input.
        """

        if quote[0] != "r":
            raise Exception(f"Invalid quote type {quote[0]}")
        quote = quote[1:]

        quote_tokens = quote.split(":")
        if len(quote_tokens) < 3:
            raise Exception(f"Quote is not compound! {quote}")

        quoteblob = base64.b64decode(quote_tokens[0])

        if compressed:
            logger.warning("Decompressing quote data which is unsafe!")
            quoteblob = zlib.decompress(quoteblob)

        qfd = -1
        quoteFile = None

        try:
            # write out quote
            qfd, qtemp = tempfile.mkstemp()
            with open(qtemp, "wb") as quoteFile:
                quoteFile.write(quoteblob)

            retDict = self.__tpm2_printquote(quoteFile.name)
            retout = retDict["retout"]
            reterr = retDict["reterr"]
            code = retDict["code"]
        except Exception as e:
            logger.error("Error printing quote: %s", str(e))
            logger.exception(e)
            return None, False
        finally:
            for fd in [qfd]:
                if fd >= 0:
                    os.close(fd)
            for fi in [quoteFile]:
                if fi is not None:
                    os.remove(fi.name)

        if len(retout) < 1 or code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
            logger.error("Failed to print quote info, output: %s", reterr)
            return None, False

        return retout, True

    def _tpm2_checkquote(self, aikTpmFromRegistrar, quote, nonce, hash_alg, compressed):
        """Write the files from data returned from tpm2_quote for running tpm2_checkquote
        :param aikTpmFromRegistrar: AIK used to generate the quote and is needed for verifying it now.
        :param quote: quote data in the format 'r<b64-compressed-quoteblob>:<b64-compressed-sigblob>:<b64-compressed-pcrblob>
        :param nonce: nonce that was used to create the quote
        :param hash_alg: the hash algorithm that was used
        :param compressed: if the quote data is compressed with zlib or not
        :returns: Returns the 'retout' from running tpm2_checkquote and True in case of success, None and False in case of error.
        This function throws an Exception on bad input.
        """
        aikFromRegistrar = tpm2_objects.pubkey_from_tpm2b_public(base64.b64decode(aikTpmFromRegistrar),).public_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PublicFormat.SubjectPublicKeyInfo,
        )

        if quote[0] != "r":
            raise Exception(f"Invalid quote type {quote[0]}")
        quote = quote[1:]

        quote_tokens = quote.split(":")
        if len(quote_tokens) < 3:
            raise Exception(f"Quote is not compound! {quote}")

        quoteblob = base64.b64decode(quote_tokens[0])
        sigblob = base64.b64decode(quote_tokens[1])
        pcrblob = base64.b64decode(quote_tokens[2])

        if compressed:
            logger.warning("Decompressing quote data which is unsafe!")
            quoteblob = zlib.decompress(quoteblob)
            sigblob = zlib.decompress(sigblob)
            pcrblob = zlib.decompress(pcrblob)

        qfd = sfd = pfd = afd = -1
        quoteFile = None
        aikFile = None
        sigFile = None
        pcrFile = None

        try:
            # write out quote
            qfd, qtemp = tempfile.mkstemp()
            with open(qtemp, "wb") as quoteFile:
                quoteFile.write(quoteblob)

            # write out sig
            sfd, stemp = tempfile.mkstemp()
            with open(stemp, "wb") as sigFile:
                sigFile.write(sigblob)

            # write out pcr
            pfd, ptemp = tempfile.mkstemp()
            with open(ptemp, "wb") as pcrFile:
                pcrFile.write(pcrblob)

            afd, atemp = tempfile.mkstemp()
            with open(atemp, "wb") as aikFile:
                aikFile.write(aikFromRegistrar)

            retDict = self.__tpm2_checkquote(aikFile.name, nonce, quoteFile.name, sigFile.name, pcrFile.name, hash_alg)
            retout = retDict["retout"]
            reterr = retDict["reterr"]
            code = retDict["code"]
        except Exception as e:
            logger.error("Error verifying quote: %s", str(e))
            logger.exception(e)
            return None, False
        finally:
            for fd in [qfd, sfd, pfd, afd]:
                if fd >= 0:
                    os.close(fd)
            for fi in [aikFile, quoteFile, sigFile, pcrFile]:
                if fi is not None:
                    os.remove(fi.name)

        if len(retout) < 1 or code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
            logger.error("Failed to validate signature, output: %s", reterr)
            return None, False

        return retout, True

    def check_quote(
        self,
        agentAttestState,
        nonce,
        data,
        quote,
        aikTpmFromRegistrar,
        tpm_policy=None,
        ima_measurement_list=None,
        allowlist=None,
        hash_alg=None,
        ima_keyrings=None,
        mb_measurement_list=None,
        mb_refstate=None,
        compressed=False,
    ) -> Failure:
        if tpm_policy is None:
            tpm_policy = {}
        if allowlist is None:
            allowlist = {}

        agent_id = agentAttestState.agent_id

        failure = Failure(Component.QUOTE_VALIDATION)
        if hash_alg is None:
            hash_alg = self.defaults["hash"]

        # First and foremost, the quote needs to be validated
        retout, success = self._tpm2_checkquote(aikTpmFromRegistrar, quote, nonce, hash_alg, compressed)
        if not success:
            # If the quote validation fails we will skip all other steps therefore this failure is irrecoverable.
            failure.add_event(
                "quote_validation", {"message": "Quote data validation using tpm2-tools", "data": retout}, False
            )
            return failure

        # Only after validating the quote, the TPM clock information can be extracted from it.
        clock_failure, current_clock_info = self.check_quote_timing(
            agentAttestState.get_tpm_clockinfo(), quote, compressed
        )
        if clock_failure:
            failure.add_event(
                "quote_validation",
                {"message": "Validation of clockinfo from quote using tpm2-tools", "data": clock_failure},
                False,
            )
            return failure
        if current_clock_info:
            agentAttestState.set_tpm_clockinfo(current_clock_info)

        pcrs = []
        jsonout = config.yaml_to_dict(retout, logger=logger)
        if jsonout is None:
            failure.add_event(
                "quote_validation",
                {"message": "YAML parsing failed for quote validation using tpm2-tools.", "data": retout},
                False,
            )
            return failure
        if "pcrs" in jsonout:
            # The hash algorithm might be in the YAML output but does not contain any data, so we also check that.
            if hash_alg in jsonout["pcrs"] and jsonout["pcrs"][hash_alg] is not None:
                alg_size = hash_alg.get_size() // 4
                for pcrval, hashval in jsonout["pcrs"][hash_alg].items():
                    pcrs.append(f"PCR {pcrval} {hashval:0{alg_size}x}")

        if len(pcrs) == 0:
            logger.warning(
                "Quote for agent %s does not contain any PCRs. Make sure that the TPM supports %s PCR banks",
                agent_id,
                str(hash_alg),
            )

        return self.check_pcrs(
            agentAttestState,
            tpm_policy,
            pcrs,
            data,
            False,
            ima_measurement_list,
            allowlist,
            ima_keyrings,
            mb_measurement_list,
            mb_refstate,
            hash_alg,
        )

    def check_quote_timing(self, previous_clockinfo, quote, compressed):
        # Sanity check quote clock information

        current_clockinfo = None

        retout, success = self._tpm2_printquote(quote, compressed)
        if not success:
            return "tpm2_print failed with " + str(retout), current_clockinfo

        tpm_data_str_dict = config.yaml_to_dict(retout, add_newlines=False, logger=logger)
        if tpm_data_str_dict is None:
            return "yaml output of tpm2_print could not be parsed!", current_clockinfo

        tentative_current_clockinfo = TPMClockInfo.from_dict(tpm_data_str_dict)

        resetdiff = tentative_current_clockinfo.resetcount - previous_clockinfo.resetcount
        restartdiff = tentative_current_clockinfo.restartcount - previous_clockinfo.restartcount

        if resetdiff < 0:
            return "resetCount value decreased on TPM between two consecutive quotes", current_clockinfo

        if restartdiff < 0:
            return "restartCount value decreased on TPM between two consecutive quotes", current_clockinfo

        if tentative_current_clockinfo.safe != 1:
            return "clock safe flag is disabled", current_clockinfo

        if not (resetdiff and restartdiff):
            if tentative_current_clockinfo.clock - previous_clockinfo.clock <= 0:
                return (
                    "clock timestamp did issued by TPM did not increase between two consecutive quotes",
                    current_clockinfo,
                )

            current_clockinfo = tentative_current_clockinfo

        return None, current_clockinfo

    def sim_extend(self, hashval_1, hashval_0=None, hash_alg=None):
        # simulate extending a PCR value by performing TPM-specific extend procedure

        if hashval_0 is None:
            hashval_0 = self.START_HASH(hash_alg)

        # compute expected value  H(0|H(data))
        extendedval = self.hashdigest(
            codecs.decode(hashval_0, "hex_codec")
            + codecs.decode(self.hashdigest(hashval_1.encode("utf-8"), hash_alg), "hex_codec"),
            hash_alg,
        ).lower()
        return extendedval

    def extendPCR(self, pcrval, hashval, hash_alg=None, lock=True):
        if hash_alg is None:
            hash_alg = self.defaults["hash"].value

        self.__run(["tpm2_pcrextend", f"{pcrval}:{str(hash_alg)}={hashval}"], lock=lock)

    def readPCR(self, pcrval, hash_alg=None):
        if hash_alg is None:
            hash_alg = self.defaults["hash"]
        if self.tools_version == "3.2":
            output = config.convert(self.__run("tpm2_pcrlist")["retout"])
        elif self.tools_version in ["4.0", "4.2", "5.4"]:
            output = config.convert(self.__run("tpm2_pcrread")["retout"])

        jsonout = config.yaml_to_dict(output, logger=logger)
        if jsonout is None:
            raise Exception("Could not read YAML output of tpm2_pcrread.")

        if not jsonout.get(hash_alg):
            raise Exception(f"Invalid hashing algorithm '{hash_alg}' for reading PCR number {pcrval}.")

        # alg_size = Hash_Algorithms.get_hash_size(hash_alg)/4
        alg_size = hash_alg.get_size() // 4
        return f"{jsonout[hash_alg][pcrval]:0{alg_size}x}"

    # tpm_random
    def _get_tpm_rand_block(self, size=32):
        # make a temp file for the output
        rand = None
        with tempfile.NamedTemporaryFile() as randpath:
            try:
                command = ["tpm2_getrandom", "-o", randpath.name, str(size)]
                retDict = self.__run(command, outputpaths=randpath.name)
                rand = retDict["fileouts"][randpath.name]
            except Exception as e:
                if not self.tpmrand_warned:
                    logger.warning("TPM randomness not available: %s", e)
                    self.tpmrand_warned = True
                return None
        return rand

    # tpm_nvram
    def write_key_nvram(self, key):
        owner_pw = self.get_tpm_metadata("owner_pw")

        # write out quote
        with tempfile.NamedTemporaryFile() as keyFile:
            keyFile.write(key)
            keyFile.flush()

            attrs = "ownerread|ownerwrite"
            # TODO(kaifeng) Escaping attrs is probably not required
            if self.tools_version == "3.2":
                self.__run(
                    [
                        "tpm2_nvdefine",
                        "-x",
                        "0x1500018",
                        "-a",
                        "0x40000001",
                        "-s",
                        str(config.BOOTSTRAP_KEY_SIZE),
                        "-t",
                        f'"{attrs}"',
                        "-I",
                        owner_pw,
                        "-P",
                        owner_pw,
                    ],
                    raiseOnError=False,
                )
                self.__run(
                    ["tpm2_nvwrite", "-x", "0x1500018", "-a", "0x40000001", "-P", owner_pw, keyFile.name],
                    raiseOnError=False,
                )
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                self.__run(
                    [
                        "tpm2_nvdefine",
                        "0x1500018",
                        "-C",
                        "0x40000001",
                        "-s",
                        str(config.BOOTSTRAP_KEY_SIZE),
                        "-a",
                        f'"{attrs}"',
                        "-p",
                        owner_pw,
                        "-P",
                        owner_pw,
                    ],
                    raiseOnError=False,
                )
                self.__run(
                    ["tpm2_nvwrite", "0x1500018", "-C", "0x40000001", "-P", owner_pw, "-i", keyFile.name],
                    raiseOnError=False,
                )

    def read_ekcert_nvram(self):
        # make a temp file for the quote
        with tempfile.NamedTemporaryFile() as nvpath:

            # Check for RSA EK cert in NVRAM (and get length)
            if self.tools_version == "3.2":
                retDict = self.__run("tpm2_nvlist", raiseOnError=False)
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                retDict = self.__run("tpm2_nvreadpublic", raiseOnError=False)
            output = retDict["retout"]
            reterr = retDict["reterr"]
            code = retDict["code"]

            if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                if self.tools_version == "3.2":
                    raise Exception("tpm2_nvlist for ekcert failed with code " + str(code) + ": " + str(reterr))
                if self.tools_version in ["4.0", "4.2", "5.4"]:
                    raise Exception("tpm2_nvreadpublic for ekcert failed with code " + str(code) + ": " + str(reterr))

            outjson = config.yaml_to_dict(output, logger=logger)

            if outjson is None or 0x1C00002 not in outjson or "size" not in outjson[0x1C00002]:
                logger.warning("No EK certificate found in TPM NVRAM")
                return None

            ekcert_size = str(outjson[0x1C00002]["size"])

            # Read the RSA EK cert from NVRAM (DER format)
            if self.tools_version == "3.2":
                retDict = self.__run(
                    ["tpm2_nvread", "-x", "0x1c00002", "-s", ekcert_size, "-f", nvpath.name, "-a", "0x01c00002"],
                    raiseOnError=False,
                    outputpaths=nvpath.name,
                )
            elif self.tools_version in ["4.0", "4.2", "5.4"]:
                retDict = self.__run(
                    ["tpm2_nvread", "0x1c00002", "-s", ekcert_size, "-o", nvpath.name],
                    raiseOnError=False,
                    outputpaths=nvpath.name,
                )
            output = config.convert(retDict["retout"])
            errout = config.convert(retDict["reterr"])
            code = retDict["code"]
            ekcert = retDict["fileouts"][nvpath.name]

            if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
                raise Exception("tpm2_nvread for ekcert failed with code " + str(code) + ": " + str(errout))

        return base64.b64encode(ekcert)

    def read_key_nvram(self):
        owner_pw = self.get_tpm_metadata("owner_pw")
        if self.tools_version == "3.2":
            retDict = self.__run(
                [
                    "tpm2_nvread",
                    "-x",
                    "0x1500018",
                    "-a",
                    "0x40000001",
                    "-s",
                    str(config.BOOTSTRAP_KEY_SIZE),
                    "-P",
                    owner_pw,
                ],
                raiseOnError=False,
            )
        elif self.tools_version in ["4.0", "4.2", "5.4"]:
            retDict = self.__run(
                ["tpm2_nvread", "0x1500018", "-C", "0x40000001", "-s", str(config.BOOTSTRAP_KEY_SIZE), "-P", owner_pw],
                raiseOnError=False,
            )

        output = retDict["retout"]
        errout = config.convert(retDict["reterr"])
        code = retDict["code"]

        if code != tpm_abstract.AbstractTPM.EXIT_SUCESS:
            if len(errout) > 0 and "handle does not exist" in "\n".join(errout):
                logger.debug("No stored U in TPM NVRAM")
                return None
            if len(errout) > 0 and "ERROR: Failed to read NVRAM public area at index" in "\n".join(errout):
                logger.debug("No stored U in TPM NVRAM")
                return None
            if len(errout) > 0 and "the handle is not correct for the use" in "\n".join(errout):
                logger.debug("No stored U in TPM NVRAM")
                return None

            raise Exception("nv_readvalue failed with code " + str(code) + ": " + str(errout))

        if len(output) != config.BOOTSTRAP_KEY_SIZE:
            logger.debug("Invalid key length from NVRAM: %d", len(output))
            return None
        return output

    @staticmethod
    def __stringify_pcr_keys(log: dict) -> None:
        """Ensure that the PCR indices are strings

        The YAML produced by `tpm2_eventlog`, when loaded by the yaml module,
        uses integer keys in the dicts holding PCR contents.  That does not
        correspond to any JSON data.  This method ensures those keys are
        strings.
        The log is untrusted because it ultimately comes from an untrusted
        source and has been processed by software that has had bugs."""
        if (not isinstance(log, dict)) or "pcrs" not in log:
            return
        old_pcrs = log["pcrs"]
        if not isinstance(old_pcrs, dict):
            return
        new_pcrs = {}
        for hash_alg, cells in old_pcrs.items():
            if not isinstance(cells, dict):
                new_pcrs[hash_alg] = cells
                continue
            new_pcrs[hash_alg] = {str(index): val for index, val in cells.items()}
        log["pcrs"] = new_pcrs
        return

    @staticmethod
    def __add_boot_aggregate(log: dict) -> None:
        """Scan the boot event log and calculate possible boot aggregates.

        Hashes are calculated for both sha1 and sha256,
        as well as for 8 or 10 participant PCRs.

        Technically the sha1/10PCR combination is unnecessary, since it has no
        implementation.

        Error conditions caused by improper string formatting etc. are
        ignored. The current assumption is that the boot event log PCR
        values are in decimal encoding, but this is liable to change."""
        if (not isinstance(log, dict)) or "pcrs" not in log:
            return
        log["boot_aggregates"] = {}
        for hashalg in log["pcrs"].keys():
            log["boot_aggregates"][hashalg] = []
            for maxpcr in [8, 10]:
                try:
                    hashclass = getattr(hashlib, hashalg)
                    h = hashclass()
                    for pcrno in range(0, maxpcr):
                        pcrstrg = log["pcrs"][hashalg][str(pcrno)]
                        pcrhex = f"{pcrstrg:0{h.digest_size*2}x}"
                        h.update(bytes.fromhex(pcrhex))
                    log["boot_aggregates"][hashalg].append(h.hexdigest())
                except Exception:
                    pass

    @staticmethod
    def __unescape_eventlog(log: dict):
        """
        Newer versions of tpm2-tools escapes the YAML output and including the trailing null byte.
        See: https://github.com/tpm2-software/tpm2-tools/commit/c78d258b2588aee535fd17594ad2f5e808056373
        This converts it back to an unescaped string.
        Example:
            '"MokList\\0"' -> 'MokList'
        """
        if tpm.tools_version in ["3.2", "4.0", "4.2"]:
            return

        escaped_chars = [
            ("\0", "\\0"),
            ("\a", "\\a"),
            ("\b", "\\b"),
            ("\t", "\\t"),
            ("\v", "\\v"),
            ("\f", "\\f"),
            ("\r", "\\r"),
            ("\x1b", "\\e"),
            ("'", "\\'"),
            ("\\", "\\\\"),
        ]

        def recursive_unescape(data):
            if isinstance(data, str):
                if data.startswith('"') and data.endswith('"'):
                    data = data[1:-1]
                    for orig, escaped in escaped_chars:
                        data = data.replace(escaped, orig)
                    data = data.rstrip("\0")
            elif isinstance(data, dict):
                for key, value in data.items():
                    data[key] = recursive_unescape(value)
            elif isinstance(data, list):
                for pos, item in enumerate(data):
                    data[pos] = recursive_unescape(item)
            return data

        recursive_unescape(log)

    def parse_binary_bootlog(self, log_bin: bytes) -> typing.Tuple[Failure, typing.Optional[dict]]:
        """Parse and enrich a BIOS boot log

        The input is the binary log.
        The output is the result of parsing and applying other conveniences."""
        failure = Failure(Component.MEASURED_BOOT, ["parser"])
        with tempfile.NamedTemporaryFile() as log_bin_file:
            log_bin_file.write(log_bin)
            log_bin_file.seek(0)
            log_bin_filename = log_bin_file.name
            try:
                retDict_tpm2 = self.__run(["tpm2_eventlog", "--eventlog-version=2", log_bin_filename])
            except Exception:
                failure.add_event("tpm2_eventlog", "running tpm2_eventlog failed", True)
                return failure, None
        log_parsed_strs = retDict_tpm2["retout"]
        if len(retDict_tpm2["reterr"]) > 0:
            failure.add_event(
                "tpm2_eventlog.warning",
                {"context": "tpm2_eventlog exited with warnings", "data": str(retDict_tpm2["reterr"])},
                True,
            )
            return failure, None
        log_parsed_data = config.yaml_to_dict(log_parsed_strs, add_newlines=False, logger=logger)
        if log_parsed_data is None:
            failure.add_event("yaml", "yaml output of tpm2_eventlog could not be parsed!", True)
            return failure, None
        # pylint: disable=import-outside-toplevel
        try:
            from keylime import tpm_bootlog_enrich
        except Exception as e:
            logger.error("Could not load tpm_bootlog_enrich (which depends on %s): %s", config.LIBEFIVAR, str(e))
            failure.add_event(
                "bootlog_enrich",
                f"Could not load tpm_bootlog_enrich (which depends on {config.LIBEFIVAR}): {str(e)}",
                True,
            )
            return failure, None
        # pylint: enable=import-outside-toplevel
        tpm_bootlog_enrich.enrich(log_parsed_data)
        tpm.__stringify_pcr_keys(log_parsed_data)
        tpm.__add_boot_aggregate(log_parsed_data)
        tpm.__unescape_eventlog(log_parsed_data)
        return failure, log_parsed_data

    def _parse_mb_bootlog(self, log_b64: str) -> typing.Tuple[Failure, typing.Optional[dict]]:
        """Parse and enrich a BIOS boot log

        The input is the base64 encoding of a binary log.
        The output is the result of parsing and applying other conveniences."""
        failure = Failure(Component.MEASURED_BOOT, ["parser"])
        try:
            log_bin = base64.b64decode(log_b64, validate=True)
            failure_mb, result = self.parse_binary_bootlog(log_bin)
            if failure_mb:
                failure.merge(failure_mb)
                result = None
        except binascii.Error:
            failure.add_event("log.base64decode", "Measured boot log could not be decoded", True)
            result = None
        return failure, result

    def parse_mb_bootlog(
        self, mb_measurement_list: str, hash_alg: algorithms.Hash
    ) -> typing.Tuple[dict, typing.Optional[dict], dict, Failure]:
        """Parse the measured boot log and return its object and the state of the PCRs
        :param mb_measurement_list: The measured boot measurement list
        :param hash_alg: the hash algorithm that should be used for the PCRs
        :returns: Returns a map of the state of the PCRs, measured boot data object and True for success
                  and False in case an error occurred
        """
        failure = Failure(Component.MEASURED_BOOT, ["parser"])
        if mb_measurement_list:
            failure_mb, mb_measurement_data = self._parse_mb_bootlog(mb_measurement_list)
            if not mb_measurement_data:
                failure.merge(failure_mb)
                logger.error("Unable to parse measured boot event log. Check previous messages for a reason for error.")
                return {}, None, {}, failure
            log_pcrs = mb_measurement_data.get("pcrs")
            if not isinstance(log_pcrs, dict):
                logger.error("Parse of measured boot event log has unexpected value for .pcrs: %r", log_pcrs)
                failure.add_event("invalid_pcrs", {"got": log_pcrs}, True)
                return {}, None, {}, failure
            pcr_hashes = log_pcrs.get(str(hash_alg))
            if (not isinstance(pcr_hashes, dict)) or not pcr_hashes:
                logger.error(
                    "Parse of measured boot event log has unexpected value for .pcrs.%s: %r", str(hash_alg), pcr_hashes
                )
                failure.add_event("invalid_pcrs_hashes", {"got": pcr_hashes}, True)
                return {}, None, {}, failure
            boot_aggregates = mb_measurement_data.get("boot_aggregates")
            if (not isinstance(boot_aggregates, dict)) or not boot_aggregates:
                logger.error(
                    "Parse of measured boot event log has unexpected value for .boot_aggragtes: %r", boot_aggregates
                )
                failure.add_event("invalid_boot_aggregates", {"got": boot_aggregates}, True)
                return {}, None, {}, failure

            return pcr_hashes, boot_aggregates, mb_measurement_data, failure

        return {}, None, {}, failure



"""Add agentstates columns

Revision ID: 7d5db1a6ffb0
Revises: f82c4252bc4f
Create Date: 2021-07-16 10:31:26.693430

"""
import sqlalchemy as sa
from alembic import op

import keylime

# revision identifiers, used by Alembic.
revision = "7d5db1a6ffb0"
down_revision = "f82c4252bc4f"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    op.add_column("verifiermain", sa.Column("boottime", sa.Integer(), nullable=True))
    op.add_column("verifiermain", sa.Column("ima_pcrs", keylime.db.verifier_db.JSONPickleType(), nullable=True))
    op.add_column("verifiermain", sa.Column("pcr10", sa.LargeBinary(), nullable=True))
    op.add_column("verifiermain", sa.Column("next_ima_ml_entry", sa.Integer(), nullable=True))


def downgrade_cloud_verifier():
    op.drop_column("verifiermain", "boottime")
    op.drop_column("verifiermain", "ima_pcrs")
    op.drop_column("verifiermain", "pcr10")
    op.drop_column("verifiermain", "next_ima_ml_entry")
